#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest import TestCase

from antlr4 import CommonTokenStream, InputStream, ParseTreeWalker
from antlr4.error.Errors import ParseCancellationException

from thespiae.path.envparse.EnvLexer import EnvLexer
from thespiae.path.envparse.EnvParserListener import EnvParserListener
from thespiae.path.factory import EnvParser
from .helper import complex_string


class _TestParserListener(EnvParserListener):

    def __init__(self):
        self._entries = []

    def enterEntry(self, ctx: EnvParser.EntryContext):
        self._entries.append(ctx.getText())

    @property
    def entries(self):
        return self._entries


def _get_token_stream(token_string: str) -> CommonTokenStream:
    lexer = EnvLexer(InputStream(token_string))
    return CommonTokenStream(lexer)


def _get_parser(token_string: str) -> EnvParser:
    return EnvParser(_get_token_stream(token_string))


class EnvLexerTest(TestCase):

    def test_string_tokenization(self):
        empty = ('', [])
        cx = (complex_string, ['"a bc"', ';', 'def', ';', "'g", ';', "h'", ';'])
        for string, tokens in [empty, cx]:
            with self.subTest(token_string=string, expected_tokens=tokens):
                stream = _get_token_stream(string)
                stream.fill()
                ts = stream.getTokens(0, len(tokens) + 1)
                self.assertSequenceEqual([t.text for t in ts], tokens)


class EnvParserTest(TestCase):

    def test_string_parsing(self):
        empty = ('', [])
        cx = (complex_string, ['"a bc"', 'def', '\'g', 'h\''])
        for string, entries in [empty, cx]:
            with self.subTest(string=string, expected_entries=entries):
                parser = _get_parser(string)
                listener = _TestParserListener()
                walker = ParseTreeWalker()
                walker.walk(listener, parser.path())
                self.assertSequenceEqual(listener.entries, entries)

    def test_path_entry_validation(self):
        self.assertEqual('abc', _get_parser('abc').single_entry().getChild(0).getText())
        with self.assertRaises(ParseCancellationException):
            _get_parser('de;f').single_entry()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from dataclasses import dataclass, field
from gettext import translation

from thespiae.conf.exception import AppEntryMixin
from thespiae.exception import ThespiaeError, generate___str__

t = translation('thespiae', fallback=True)
_ = t.gettext


class PathEntryError(ThespiaeError):
    pass


class PathError(ThespiaeError):
    pass


class InconsistentUserPathEntryDataError(ThespiaeError):
    pass


class InconsistentActualPathEntryDataError(ThespiaeError):
    pass


@generate___str__(_('Invalid value of system PATH'))
@dataclass(frozen=True)
class InvalidSystemPathError(ThespiaeError):
    pass


@generate___str__(_('Invalid value of user PATH'))
@dataclass(frozen=True)
class InvalidUserPathError(ThespiaeError):
    pass


@generate___str__
@dataclass(frozen=True)
class _InconsistentDataExtensionMixin:
    data: str = field(metadata={'format': _('raw string {}')})
    extended_data: str = field(metadata={'format': _('extended string {}')})


@generate___str__(_('Number of user path entries is not preserved with regard to extension'))
@dataclass(frozen=True)
class InconsistentUserPathExtensionError(_InconsistentDataExtensionMixin, ThespiaeError):
    pass


@generate___str__(_('Invalid path entry'))
@dataclass(frozen=True)
class InvalidPathEntryError(AppEntryMixin, ThespiaeError):
    entry: str = field(metadata={'format': _('path element string {}')})


@generate___str__(_('Path element is not consistent with regard to extension'))
@dataclass(frozen=True)
class InconsistentPathEntryExtensionError(_InconsistentDataExtensionMixin, AppEntryMixin, ThespiaeError):
    pass



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from string import Template
from typing import TYPE_CHECKING

from .singleton import path_manager

if TYPE_CHECKING:
    from typing import Mapping

complex_string = ' "a bc"; def;\'g;h\';'


class PathManagerMockResetMixin:

    def setUp(self, *args, **kwarg):
        super().setUp(*args, **kwarg)
        path_manager.reset_mock(return_value=True, side_effect=True)
        path_manager.get_system_path.reset_mock(return_value=True, side_effect=True)
        path_manager.get_user_path.reset_mock(return_value=True, side_effect=True)
        path_manager.extend_path_data.reset_mock(return_value=True, side_effect=True)


class ExtensionTemplate(Template):
    delimiter = ''
    idpattern = '%[^=%]+%'


def _helper(data: str, path_extensions: Mapping[str, str]):
    prev = ''
    while prev != data:
        prev = data
        data = ExtensionTemplate(data).safe_substitute(path_extensions)
    return data


def set_path_manager_data(full_extended_path: str, user_path: str, path_extensions: Mapping[str, str]):
    path_manager.get_system_path.return_value = full_extended_path
    path_manager.get_user_path.return_value = user_path
    path_extensions = {'%{}%'.format(key): path_extensions[key] for key in path_extensions}
    path_manager.extend_path_data.side_effect = lambda arg: _helper(arg, path_extensions)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from thespiae.conf import AppEntry

if TYPE_CHECKING:
    from typing import Optional


@dataclass(frozen=True)
class DownloadSpec(AppEntry):
    url: str = None
    hash: str = None
    download_path: str = None
    part_path: str = None


@dataclass(frozen=True)
class ContentRangeData:
    start: Optional[int]
    end: Optional[int]
    total: Optional[int]



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from typing import Sequence

    from thespiae.conf import AppEntry
    from .data import DownloadSpec


class Feedback(Protocol):

    @abstractmethod
    def report_checking_software(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def confirm_operations(self, to_download: Sequence[DownloadSpec], to_uninstall: Sequence[AppEntry],
                           to_install: Sequence[AppEntry]) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_software_set_no_changes(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_download_started(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_entry_download_initiated(self, spec: DownloadSpec) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_entry_download_started(self, spec: DownloadSpec, current_size: int, total_size: int = None) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_entry_download_progress(self, spec: DownloadSpec, batch_size: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_entry_download_finished(self, spec: DownloadSpec) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_download_finished(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_removal_started(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_entry_removal_started(self, entry: AppEntry) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_entry_removal_finished(self, entry: AppEntry) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_removal_finished(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_installation_started(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_entry_installation_started(self, entry: AppEntry) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_entry_installation_finished(self, entry: AppEntry) -> None:
        raise NotImplementedError

    @abstractmethod
    def report_installation_finished(self) -> None:
        raise NotImplementedError



# Generated from EnvParser.g4 by ANTLR 4.7.2
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .EnvParser import EnvParser
else:
    from EnvParser import EnvParser


# This class defines a complete listener for a parse tree produced by EnvParser.
class EnvParserListener(ParseTreeListener):

    # Enter a parse tree produced by EnvParser#entry.
    def enterEntry(self, ctx: EnvParser.EntryContext):
        pass

    # Exit a parse tree produced by EnvParser#entry.
    def exitEntry(self, ctx: EnvParser.EntryContext):
        pass

    # Enter a parse tree produced by EnvParser#path.
    def enterPath(self, ctx: EnvParser.PathContext):
        pass

    # Exit a parse tree produced by EnvParser#path.
    def exitPath(self, ctx: EnvParser.PathContext):
        pass

    # Enter a parse tree produced by EnvParser#single_entry.
    def enterSingle_entry(self, ctx: EnvParser.Single_entryContext):
        pass

    # Exit a parse tree produced by EnvParser#single_entry.
    def exitSingle_entry(self, ctx: EnvParser.Single_entryContext):
        pass



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from thespiae.cli import CLI
from thespiae.conf import parse_config_file
from thespiae.install import handle_software_changes
from thespiae.path import update_user_path


def main():
    with CLI('thespiae') as ui:
        config_data = parse_config_file(ui.params.config_file)
        handle_software_changes(ui.params.data_dir, config_data, ui.feedback)
        update_user_path(config_data, ui.feedback)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from ctypes.wintypes import HWND, UINT
from unittest import TestCase
from unittest.mock import patch, DEFAULT
from winreg import HKEY_CURRENT_USER, KEY_SET_VALUE, REG_EXPAND_SZ, HKEY_LOCAL_MACHINE

import thespiae.path.system as ts
from .singleton import real_path_manager as path_manager


class PathManagerTestCase(TestCase):

    @patch.multiple(ts, OpenKeyEx=DEFAULT, QueryValueEx=DEFAULT, CloseKey=DEFAULT)
    def test_getting_system_path(self, OpenKeyEx, QueryValueEx, CloseKey):
        QueryValueEx.return_value = 'test_system_path', None
        self.assertEqual(path_manager.get_system_path(), 'test_system_path')
        OpenKeyEx.assert_called_once_with(HKEY_LOCAL_MACHINE,
                                          r'System\CurrentControlSet\Control\Session Manager\Environment')
        QueryValueEx.assert_called_once_with(OpenKeyEx.return_value, 'Path')
        CloseKey.assert_called_once_with(OpenKeyEx.return_value)

    @patch.multiple(ts, OpenKeyEx=DEFAULT, QueryValueEx=DEFAULT, CloseKey=DEFAULT)
    def test_getting_user_path(self, OpenKeyEx, QueryValueEx, CloseKey):
        QueryValueEx.return_value = 'test_user_path', None
        self.assertEqual(path_manager.get_user_path(), 'test_user_path')
        OpenKeyEx.assert_called_once_with(HKEY_CURRENT_USER, 'Environment')
        QueryValueEx.assert_called_once_with(OpenKeyEx.return_value, 'Path')
        CloseKey.assert_called_once_with(OpenKeyEx.return_value)

    @patch.multiple(ts, ExpandEnvironmentStrings=DEFAULT)
    def test_extending_path_data(self, ExpandEnvironmentStrings):
        ExpandEnvironmentStrings.return_value = 'extended_test_data'
        self.assertEqual(path_manager.extend_path_data('test_data'), 'extended_test_data')
        ExpandEnvironmentStrings.assert_called_once_with('test_data')

    @patch.multiple(ts, OpenKeyEx=DEFAULT, CloseKey=DEFAULT, SetValueEx=DEFAULT)
    def test_setting_user_path(self, OpenKeyEx, CloseKey, SetValueEx):
        path_manager.set_user_path('test_user_path')
        OpenKeyEx.assert_called_once_with(HKEY_CURRENT_USER, 'Environment', access=KEY_SET_VALUE)
        SetValueEx.assert_called_once_with(OpenKeyEx.return_value, 'Path', 0, REG_EXPAND_SZ, 'test_user_path')
        CloseKey.assert_called_once_with(OpenKeyEx.return_value)

    @patch.multiple(ts, WinDLL=DEFAULT)
    def test_notifying_path_changes(self, WinDLL):
        path_manager.notify_path_change()
        WinDLL.assert_called_once_with('user32', use_last_error=True)
        args = WinDLL.return_value.SendMessageTimeoutW.call_args
        hwnd, msg, wParam, lParam, fuFlags, uTimeout, lpdwResult = args[0]
        self.assertEqual(hwnd.value, HWND(0xffff).value)
        self.assertEqual(msg.value, UINT(0x001A).value)
        self.assertIsNone(wParam)
        self.assertEqual(lParam.value, 'Environment')
        self.assertEqual(fuFlags.value, UINT(0x0002 & 0x0008).value)
        self.assertEqual(uTimeout.value, UINT(1000).value)
        self.assertIsNone(lpdwResult)

    @patch.multiple(ts, WinDLL=DEFAULT)
    def test_notifying_path_changes2(self, WinDLL):
        WinDLL.return_value.SendMessageTimeoutW.return_value = 0
        with self.assertRaises(OSError):
            path_manager.notify_path_change()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import builtins
import sys
from unittest import TestCase
from unittest.mock import patch, DEFAULT, NonCallableMagicMock, call, ANY

import thespiae.cli
from thespiae.conf import AppEntry
from thespiae.exception import ThespiaeError
from thespiae.install import DownloadSpec


@patch.object(sys, 'argv', new=['script_path', 'config_file_path', 'data_dir_path'])
@patch.multiple(thespiae.cli, _error=DEFAULT, _position_after_bars=DEFAULT)
class CLITestCase(TestCase):

    def test_param_parsing(self, _error, _position_after_bars):
        ui = thespiae.cli.CLI('thespiae')
        self.assertEqual(ui.params.config_file, 'config_file_path')
        self.assertEqual(ui.params.data_dir, 'data_dir_path')
        _error.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_keyboard_interrupt(self, _error, _position_after_bars):
        with thespiae.cli.CLI('thespiae'):
            raise KeyboardInterrupt
        _error.assert_called()
        _position_after_bars.assert_called_once()

    def test_no_confirm_interrupt(self, _error, _position_after_bars):
        with thespiae.cli.CLI('thespiae'):
            raise thespiae.cli._NotConfirmedError
        _error.assert_called()
        _position_after_bars.assert_called_once()

    def test_application_error(self, _error, _position_after_bars):
        class TestError(ThespiaeError):

            def __str__(self):
                return '123'

        with thespiae.cli.CLI('thespiae'):
            raise TestError
        _error.assert_called_once_with('123')
        _position_after_bars.assert_called_once()


@patch.object(sys, 'argv', new=['script_path', 'config_file_path', 'data_dir_path'])
@patch.multiple(thespiae.cli, _error=DEFAULT, _info=DEFAULT, _confirm=DEFAULT, tqdm=DEFAULT,
                _position_after_bars=DEFAULT)
class CLIFeedbackTestCase(TestCase):

    def test_report_checking_software(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_checking_software()
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_confirm_operations(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.confirm_operations([DownloadSpec(name='test')], [AppEntry(name='test2')],
                                           [AppEntry(name='test3')])
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_called_once()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_software_set_no_changes(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_software_set_no_changes()
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_download_started(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_download_started()
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_entry_download_initiated(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.confirm_operations([DownloadSpec(name='test')], [AppEntry(name='test2')],
                                           [AppEntry(name='test3')])
            ui.feedback.report_entry_download_initiated(DownloadSpec(name='test'))
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_called()
        tqdm.assert_called_once()
        _position_after_bars.assert_not_called()

    def test_report_entry_download_started(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.confirm_operations([DownloadSpec(name='test')], [AppEntry(name='test2')],
                                           [AppEntry(name='test3')])
            ui.feedback.report_entry_download_initiated(DownloadSpec(name='test'))
            ui.feedback.report_entry_download_started(DownloadSpec(name='test'), 5, 15)
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_called_once()
        tqdm.assert_called_once()
        self.assertEqual(tqdm.return_value.total, 15)
        self.assertEqual(tqdm.return_value.n, 5)
        tqdm.return_value.refresh.assert_called_once()
        _position_after_bars.assert_not_called()

    def test_report_entry_download_progress(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.confirm_operations([DownloadSpec(name='test')], [AppEntry(name='test2')],
                                           [AppEntry(name='test3')])
            ui.feedback.report_entry_download_initiated(DownloadSpec(name='test'))
            ui.feedback.report_entry_download_progress(DownloadSpec(name='test'), 10)
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_called_once()
        tqdm.assert_called_once()
        tqdm.return_value.update.assert_called_once_with(10)
        _position_after_bars.assert_not_called()

    def test_report_entry_download_finished(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_entry_download_finished(DownloadSpec(name='test'))
        _info.assert_not_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_download_finished(self, _error, _info, _confirm, tqdm, _position_after_bars):
        items = [NonCallableMagicMock(), NonCallableMagicMock()]
        tqdm.side_effect = items
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.confirm_operations([DownloadSpec(name='test'), DownloadSpec(name='test2')], [], [])
            ui.feedback.report_entry_download_initiated(DownloadSpec(name='test'))
            ui.feedback.report_entry_download_initiated(DownloadSpec(name='test2'))
            ui.feedback.report_entry_download_started(DownloadSpec(name='test'), 5)
            ui.feedback.report_entry_download_started(DownloadSpec(name='test2'), 10)
            ui.feedback.report_entry_download_progress(DownloadSpec(name='test'), 5)
            ui.feedback.report_entry_download_progress(DownloadSpec(name='test2'), 10)
            ui.feedback.report_download_finished()
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_called()
        self.assertEqual(tqdm.call_count, 2)
        _position_after_bars.assert_called_once_with(2)
        self.assertSequenceEqual(items[0].method_calls, [call.refresh(), call.update(5), call.refresh(), call.close()])
        self.assertSequenceEqual(items[1].method_calls, [call.refresh(), call.update(10), call.refresh(), call.close()])

    def test_report_removal_started(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_removal_started()
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_entry_removal_started(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_entry_removal_started(AppEntry(name='test2'))
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_entry_removal_finished(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_entry_removal_finished(AppEntry(name='test2'))
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_removal_finished(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_removal_finished()
        _info.assert_not_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_installation_started(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_installation_started()
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_entry_installation_started(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_entry_installation_started(AppEntry(name='test2'))
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_entry_installation_finished(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_entry_installation_finished(AppEntry(name='test2'))
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_installation_finished(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_installation_finished()
        _info.assert_not_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_path_analysis(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_path_analysis()
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_confirm_user_path_update(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.confirm_user_path_update('before', 'after')
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_called_once()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_user_path_no_changes(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_user_path_no_changes()
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()

    def test_report_user_path_updated(self, _error, _info, _confirm, tqdm, _position_after_bars):
        with thespiae.cli.CLI('thespiae') as ui:
            ui.feedback.report_user_path_updated()
        _info.assert_called()
        _error.assert_not_called()
        _confirm.assert_not_called()
        tqdm.assert_not_called()
        _position_after_bars.assert_not_called()


@patch.multiple(builtins, print=DEFAULT, input=DEFAULT)
class InternalCLITestCase(TestCase):

    def test_displaying_info(self, print, input):
        thespiae.cli._info('test_data')
        print.assert_called_once_with('test_data', file=ANY)
        input.assert_not_called()

    def test_displaying_error(self, print, input):
        thespiae.cli._error('test_data')
        print.assert_called_once_with('test_data', file=ANY)
        input.assert_not_called()

    def test_asking_confirmation(self, print, input):
        input.return_value = 'y'
        thespiae.cli._confirm('test_data')
        input.assert_called_once()
        print.assert_not_called()

    @patch.multiple(sys, stdout=DEFAULT)
    def test_moving_cursor(self, stdout, print, input):
        thespiae.cli._position_after_bars(5)
        self.assertSequenceEqual(stdout.mock_calls, [call.write('\r'), call.write('\x1b[4B'), call.flush()])
        print.assert_not_called()
        input.assert_not_called()

    @patch.multiple(sys, stdout=DEFAULT)
    def test_moving_cursor2(self, stdout, print, input):
        thespiae.cli._position_after_bars(0)
        self.assertSequenceEqual(stdout.mock_calls, [])
        print.assert_not_called()
        input.assert_not_called()

    def test_asking_confirmation2(self, print, input):
        input.return_value = ''
        with self.assertRaises(thespiae.cli._NotConfirmedError):
            thespiae.cli._confirm('test_data')
        input.assert_called_once()
        print.assert_not_called()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from dataclasses import fields
from functools import singledispatch, partial
from typing import List, Mapping, Set

META_FORMAT = 'format'


@singledispatch
def generate___str__(message: str):
    return partial(_generate___str__, message=message)


def _format_helper(obj) -> str:
    if isinstance(obj, Mapping):
        return ', '.join((f'{_format_helper(key)}:{_format_helper(value)}' for key, value in obj.items()))
    elif isinstance(obj, (List, Set)):
        return ', '.join((f'{_format_helper(item)}' for item in obj))
    else:
        return obj


@generate___str__.register
def _generate___str__(cls: type, message: str = None):
    if '__str__' in cls.__dict__:
        raise TypeError(f'{cls.__name__} explicitly defines __str__')

    def __str__(self):
        nonlocal message
        if message:
            es = [message]
        else:
            es = []
        fs = []
        for field in fields(cls):
            fmt = field.metadata.get(META_FORMAT, None)
            if fmt:
                val = getattr(self, field.name, None)
                if val is not None:
                    fs.append(fmt.format(_format_helper(val)))
        if fs:
            es.append(', '.join(fs))
        return '; '.join(es)

    setattr(cls, '__str__', __str__)
    return cls


class ThespiaeError(Exception):
    pass



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from typing import TYPE_CHECKING

from antlr4 import CommonTokenStream, ParseTreeWalker, InputStream
from antlr4.error.Errors import ParseCancellationException

from .envparse.EnvLexer import EnvLexer
from .envparse.EnvParserListener import EnvParserListener
from .exception import PathEntryError, PathError, InconsistentUserPathEntryDataError, \
    InconsistentActualPathEntryDataError, InconsistentUserPathExtensionError, InconsistentPathEntryExtensionError, \
    InvalidPathEntryError, InvalidSystemPathError, InvalidUserPathError
from .factory import EnvParser

if TYPE_CHECKING:
    from typing import List, Container, Sequence, Tuple
    
    from thespiae.conf.data import AppData, AppEntry
    from .protocol import Feedback
    from .system import PathManager


class _EntryListener(EnvParserListener):

    def __init__(self):
        self._processed_entries = []
        self._raw_entries = []

    def enterEntry(self, ctx: EnvParser.EntryContext) -> None:
        ns = ctx.UNQUOTED_STRING()
        if ns:
            raw = processed = ns.getText()
        else:
            raw = ctx.DOUBLE_QUOTED_STRING().getText()
            processed = raw[1:-1]
        self._raw_entries.append(raw)
        self._processed_entries.append(processed)

    @property
    def raw_entries(self) -> List[str]:
        return self._raw_entries

    @property
    def processed_entries(self) -> List[str]:
        return self._processed_entries


def _parser(path: str, build_trees=True) -> EnvParser:
    return EnvParser(CommonTokenStream(EnvLexer(InputStream(path))), build_parse_trees=build_trees)


def get_path_entries(path_string: str, raw: bool) -> List[str]:
    walker = ParseTreeWalker()
    ls = _EntryListener()
    p = _parser(path_string)
    try:
        walker.walk(ls, p.path())
    except ParseCancellationException:
        raise PathError
    return ls.raw_entries if raw else ls.processed_entries


def get_single_path_entry(data: str, raw: bool) -> str:
    walker = ParseTreeWalker()
    ls = _EntryListener()
    try:
        walker.walk(ls, _parser(data).single_entry())
    except ParseCancellationException:
        raise PathEntryError
    return ls.raw_entries[0] if raw else ls.processed_entries[0]


def construct_new_user_path(system_path_entries: Container[str], raw_user_path_entries: Sequence[str],
                            user_path_entries: Sequence[str], raw_actual_entries: Sequence[str],
                            actual_entries: Sequence[str], obsolete_entries: Container[str]):
    new_user_path_entries = []
    if len(raw_user_path_entries) != len(user_path_entries):
        raise InconsistentUserPathEntryDataError
    if len(raw_actual_entries) != len(actual_entries):
        raise InconsistentActualPathEntryDataError
    for raw_entry, entry in zip(raw_user_path_entries, user_path_entries):
        if entry in actual_entries or entry not in obsolete_entries:
            new_user_path_entries.append(raw_entry)
    for raw_entry, entry in zip(raw_actual_entries, actual_entries):
        if entry not in system_path_entries and entry not in user_path_entries:
            new_user_path_entries.append(raw_entry)
    return ';'.join(new_user_path_entries)


class PathProcessor:

    def __init__(self, manager: PathManager):
        self.manager = manager

    def _extract_path_data(self, entry: AppEntry) -> Tuple[List[str], List[str]]:
        single_raw_entries = []
        single_entries = []
        if entry.path_entries:
            for path_entry in entry.path_entries:
                try:
                    single_raw_entry = get_single_path_entry(path_entry, True)
                except PathEntryError:
                    raise InvalidPathEntryError(entry, path_entry)
                else:
                    single_raw_entries.append(single_raw_entry)
                    extended_entry = self.manager.extend_path_data(single_raw_entry)
                    try:
                        single_entry = get_single_path_entry(extended_entry, False)
                    except PathEntryError:
                        raise InconsistentPathEntryExtensionError(entry, path_entry, extended_entry)
                    else:
                        single_entries.append(single_entry)
        return single_raw_entries, single_entries

    def process_path_changes(self, config: AppData, fb: Feedback) -> None:
        raw_actual_entries = []
        actual_entries = []
        raw_obsolete_entries = []
        obsolete_entries = []
        fb.report_path_analysis()
        for e in config.to_install:
            raw, extended = self._extract_path_data(e)
            raw_actual_entries += raw
            actual_entries += extended
        for e in config.to_uninstall:
            raw, extended = self._extract_path_data(e)
            raw_obsolete_entries += raw
            obsolete_entries += extended
        try:
            system = get_path_entries(self.manager.extend_path_data(self.manager.get_system_path()), False)
        except PathError:
            raise InvalidSystemPathError
        else:
            raw_user_path = self.manager.get_user_path()
            try:
                raw_user_path_entries = get_path_entries(raw_user_path, True)
                user_path = self.manager.extend_path_data(raw_user_path)
                user_path_entries = get_path_entries(user_path, False)
            except PathError:
                raise InvalidUserPathError
            else:
                try:
                    new_user_path = construct_new_user_path(set(system), raw_user_path_entries, user_path_entries,
                                                            raw_actual_entries, actual_entries, set(obsolete_entries))
                except InconsistentUserPathEntryDataError:
                    raise InconsistentUserPathExtensionError(raw_user_path, user_path)
                else:
                    if new_user_path != raw_user_path:
                        fb.confirm_user_path_update(raw_user_path, new_user_path)
                        self.manager.set_user_path(new_user_path)
                        self.manager.notify_path_change()
                        fb.report_user_path_updated()
                    else:
                        fb.report_user_path_no_changes()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from dataclasses import replace
from pathlib import Path
from unittest import TestCase
from unittest.mock import call, NonCallableMock

from thespiae.conf import AppData, AppEntry
from thespiae.conf.core import get_app_config_from
from thespiae.install.exception import InvalidOrMissingAppDataError, InterruptedFileOperationsError, \
    UnknownInstallTypeError, UnknownUninstallTypeError
from thespiae.install.protocol import Feedback
from .helper import set_install_manager_data, InstallManagerMockResetMixin, DownloadManagerMockResetMixin, \
    ExtraAssertMixin, set_download_manager_data
from .singleton import exe_install_handler, msi_install_handler, command_install_handler, install_manager, \
    software_processor, download_manager, exe_uninstall_handler, msi_uninstall_handler, command_uninstall_handler, \
    file_install_handler, file_uninstall_handler, archive_install_handler, archive_uninstall_handler
from ..conf.helper import load_yaml_from_test_dir

_exe_entry1 = AppEntry(name='test', installer_url='http://example.com/123.exe', file_hash='4567', version='890',
                       uninstaller_path='123.exe', install_args=['a'], uninstall_args=['b'])


class ExeInstallHandlerTest(InstallManagerMockResetMixin, TestCase):

    def test_applicability(self):
        self.assertTrue(exe_install_handler.is_applicable(_exe_entry1))

    def test_applicability2(self):
        self.assertFalse(exe_install_handler.is_applicable(replace(_exe_entry1, installer_url=None)))

    def test_creating_download_spec(self):
        spec = exe_install_handler.create_download_spec(_exe_entry1, 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/123.exe')
        self.assertEqual(spec.download_path, 'C:\\test\\890\\test_890.exe')
        self.assertEqual(spec.hash, '4567')
        self.assertEqual(spec.name, 'test')
        self.assertEqual(spec.version, '890')

    def test_creating_download_spec2(self):
        e = replace(_exe_entry1, name='')
        with self.assertRaises(InvalidOrMissingAppDataError) as c:
            exe_install_handler.create_download_spec(e, 'C:\\')
        self.assertEqual(c.exception.app_entry, e)
        self.assertEqual(c.exception.key, 'name')
        self.assertEqual(c.exception.value, '')

    def test_creating_download_spec3(self):
        spec = exe_install_handler.create_download_spec(replace(_exe_entry1, version=None), 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/123.exe')
        self.assertEqual(spec.download_path, 'C:\\test\\test.exe')
        self.assertEqual(spec.hash, '4567')
        self.assertEqual(spec.name, 'test')

    def test_creating_download_spec4(self):
        spec = exe_install_handler.create_download_spec(replace(_exe_entry1, architecture='x86_64'), 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/123.exe')
        self.assertEqual(spec.download_path, 'C:\\test\\890\\x86_64\\test_890_x86_64.exe')
        self.assertEqual(spec.hash, '4567')
        self.assertEqual(spec.name, 'test')
        self.assertEqual(spec.version, '890')
        self.assertEqual(spec.architecture, 'x86_64')

    def test_installing(self):
        ds = exe_install_handler.create_download_spec(_exe_entry1, 'C:\\')
        exe_install_handler.install(_exe_entry1, ds)
        install_manager.run_file.assert_called_once_with('C:\\test\\890\\test_890.exe', ['a'])


class ExeUninstallHandlerTest(InstallManagerMockResetMixin, TestCase):

    def test_applicability(self):
        self.assertTrue(exe_uninstall_handler.is_applicable(_exe_entry1))

    def test_applicability2(self):
        self.assertFalse(exe_uninstall_handler.is_applicable(replace(_exe_entry1, uninstaller_path=None)))

    def test_is_installed_checking(self):
        set_install_manager_data(True, None, None)
        self.assertTrue(exe_uninstall_handler.is_installed(_exe_entry1))
        install_manager.is_file_present.assert_called_once_with('123.exe')

    def test_uninstalling(self):
        exe_uninstall_handler.uninstall(_exe_entry1)
        install_manager.run_file.assert_called_once_with('123.exe', ['b'])


_msi_entry1 = AppEntry(name='test', version='890', package_url='http://example.com/123.msi', product_code='123',
                       file_hash='4567', install_args=['a'], uninstall_args=['b'])


class MSIInstallHandlerTest(InstallManagerMockResetMixin, TestCase):

    def test_applicability(self):
        self.assertTrue(msi_install_handler.is_applicable(_msi_entry1))

    def test_applicability2(self):
        self.assertFalse(msi_install_handler.is_applicable(replace(_msi_entry1, package_url=None)))

    def test_creating_download_spec(self):
        spec = msi_install_handler.create_download_spec(_msi_entry1, 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/123.msi')
        self.assertEqual(spec.download_path, 'C:\\test\\890\\test_890.msi')
        self.assertEqual(spec.hash, '4567')
        self.assertEqual(spec.name, 'test')
        self.assertEqual(spec.version, '890')

    def test_creating_download_spec2(self):
        e = replace(_msi_entry1, name='')
        with self.assertRaises(InvalidOrMissingAppDataError) as c:
            msi_install_handler.create_download_spec(e, 'C:\\')
        self.assertEqual(c.exception.app_entry, e)
        self.assertEqual(c.exception.key, 'name')
        self.assertEqual(c.exception.value, '')

    def test_creating_download_spec3(self):
        spec = msi_install_handler.create_download_spec(replace(_msi_entry1, version=None), 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/123.msi')
        self.assertEqual(spec.download_path, 'C:\\test\\test.msi')
        self.assertEqual(spec.hash, '4567')
        self.assertEqual(spec.name, 'test')

    def test_creating_download_spec4(self):
        spec = msi_install_handler.create_download_spec(replace(_msi_entry1, architecture='x86_64'), 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/123.msi')
        self.assertEqual(spec.download_path, 'C:\\test\\890\\x86_64\\test_890_x86_64.msi')
        self.assertEqual(spec.hash, '4567')
        self.assertEqual(spec.name, 'test')
        self.assertEqual(spec.version, '890')
        self.assertEqual(spec.architecture, 'x86_64')

    def test_installing(self):
        msi_install_handler.install(_msi_entry1, msi_install_handler.create_download_spec(_msi_entry1, 'C:\\'))
        install_manager.install_msi.assert_called_once_with('C:\\test\\890\\test_890.msi', ['a'])


class MSIUninstallHandlerTest(InstallManagerMockResetMixin, TestCase):

    def test_applicability(self):
        self.assertTrue(msi_uninstall_handler.is_applicable(_msi_entry1))

    def test_applicability2(self):
        self.assertFalse(msi_uninstall_handler.is_applicable(replace(_msi_entry1, product_code=None)))

    def test_is_installed_checking(self):
        set_install_manager_data(None, True, None)
        self.assertTrue(msi_uninstall_handler.is_installed(_msi_entry1))
        install_manager.is_product_code_present.assert_called_once_with('123')

    def test_uninstalling(self):
        msi_uninstall_handler.uninstall(_msi_entry1)
        install_manager.uninstall_msi.assert_called_once_with('123', ['b'])


_command_entry1 = AppEntry(name='test', version='456', command='testc', install_args=['a'], list_args=['b'],
                           installed_list_entry='123:456', uninstall_args=['c'])


class CommandInstallHandlerTest(InstallManagerMockResetMixin, TestCase):

    def test_applicability(self):
        self.assertTrue(command_install_handler.is_applicable(_command_entry1))

    def test_applicability2(self):
        self.assertFalse(command_install_handler.is_applicable(replace(_msi_entry1, command=None)))

    def test_creating_download_spec(self):
        self.assertIsNone(command_install_handler.create_download_spec(_msi_entry1, 'C:\\'))

    def test_installing(self):
        command_install_handler.install(_command_entry1, None)
        install_manager.run_command.assert_called_once_with('testc', ['a'])


class CommandUninstallHandlerTest(InstallManagerMockResetMixin, TestCase):

    def test_applicability(self):
        self.assertTrue(command_uninstall_handler.is_applicable(_command_entry1))

    def test_applicability2(self):
        self.assertFalse(command_uninstall_handler.is_applicable(replace(_command_entry1, list_args=None)))

    def test_applicability3(self):
        self.assertFalse(command_uninstall_handler.is_applicable(replace(_command_entry1, installed_list_entry=None)))

    def test_is_installed_checking(self):
        set_install_manager_data(None, None, 'a:b\r\n123:456\r\nc:d\r\n')
        self.assertTrue(command_uninstall_handler.is_installed(_command_entry1))
        install_manager.run_command.assert_called_once_with('testc', ['b'])

    def test_uninstalling(self):
        command_uninstall_handler.uninstall(_command_entry1)
        install_manager.run_command.assert_called_once_with('testc', ['c'])


_file_entry1 = AppEntry(name='123', version='456', file_url='http://example.com/1.txt',
                        file_directory=r'C:\test\test2', file_name='test.txt', file_hash='789')


class FileInstallTestCase(InstallManagerMockResetMixin, TestCase):

    def test_applicability(self):
        self.assertTrue(file_install_handler.is_applicable(_file_entry1))

    def test_applicability2(self):
        self.assertFalse(file_install_handler.is_applicable(replace(_file_entry1, file_url=None)))

    def test_applicability3(self):
        self.assertFalse(file_install_handler.is_applicable(replace(_file_entry1, file_directory=None)))

    def test_applicability4(self):
        self.assertFalse(file_install_handler.is_applicable(replace(_file_entry1, file_name=None)))

    def test_creating_download_spec(self):
        spec = file_install_handler.create_download_spec(_file_entry1, 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/1.txt')
        self.assertEqual(spec.download_path, 'C:\\123\\456\\123_456.txt')
        self.assertEqual(spec.part_path, 'C:\\123\\456\\123_456.txt.part')
        self.assertEqual(spec.hash, '789')
        self.assertEqual(spec.name, '123')
        self.assertEqual(spec.version, '456')

    def test_creating_download_spec2(self):
        spec = file_install_handler.create_download_spec(replace(_file_entry1, version=None), 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/1.txt')
        self.assertEqual(spec.download_path, 'C:\\123\\123.txt')
        self.assertEqual(spec.part_path, 'C:\\123\\123.txt.part')
        self.assertEqual(spec.hash, '789')
        self.assertEqual(spec.name, '123')

    def test_creating_download_spec3(self):
        spec = file_install_handler.create_download_spec(replace(_file_entry1, architecture='x86_64'), 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/1.txt')
        self.assertEqual(spec.download_path, 'C:\\123\\456\\x86_64\\123_456_x86_64.txt')
        self.assertEqual(spec.part_path, 'C:\\123\\456\\x86_64\\123_456_x86_64.txt.part')
        self.assertEqual(spec.hash, '789')
        self.assertEqual(spec.name, '123')
        self.assertEqual(spec.version, '456')

    def test_installing(self):
        file_install_handler.install(_file_entry1, file_install_handler.create_download_spec(_file_entry1, 'C:\\'))
        install_manager.copy_file.assert_called_once_with('C:\\123\\456\\123_456.txt', Path('C:/test/test2'),
                                                          'test.txt')

    def test_installing2(self):
        e = replace(_file_entry1, file_directory=[])
        with self.assertRaises(InvalidOrMissingAppDataError) as c:
            file_install_handler.install(e, file_install_handler.create_download_spec(e, 'C:\\'))
        self.assertEqual(c.exception.app_entry, e)
        self.assertEqual(c.exception.key, 'file_directory')
        self.assertEqual(c.exception.value, [])

    def test_installing4(self):
        install_manager.copy_file.side_effect = OSError
        with self.assertRaises(InterruptedFileOperationsError) as c:
            file_install_handler.install(_file_entry1, file_install_handler.create_download_spec(_file_entry1, 'C:\\'))
        self.assertEqual(c.exception.app_entry, _file_entry1)
        self.assertEqual(c.exception.root_directory, r'C:\test\test2')


class FileUninstallTestCase(InstallManagerMockResetMixin, TestCase):

    def test_applicability(self):
        self.assertTrue(file_uninstall_handler.is_applicable(_file_entry1))

    def test_applicability2(self):
        self.assertFalse(file_uninstall_handler.is_applicable(replace(_file_entry1, file_directory=None)))

    def test_applicability3(self):
        self.assertFalse(file_uninstall_handler.is_applicable(replace(_file_entry1, file_name=None)))

    def test_is_installed(self):
        file_uninstall_handler.is_installed(_file_entry1)
        install_manager.is_file_present.assert_called_once_with('C:\\test\\test2\\test.txt')

    def test_uninstalling(self):
        file_uninstall_handler.uninstall(_file_entry1)
        install_manager.remove_file.assert_called_once_with(Path('C:/test/test2'), 'test.txt')

    def test_uninstalling2(self):
        e = replace(_file_entry1, file_directory=[])
        with self.assertRaises(InvalidOrMissingAppDataError) as c:
            file_uninstall_handler.uninstall(e)
        self.assertEqual(c.exception.app_entry, e)
        self.assertEqual(c.exception.key, 'file_directory')
        self.assertEqual(c.exception.value, [])

    def test_uninstalling4(self):
        install_manager.remove_file.side_effect = OSError
        with self.assertRaises(InterruptedFileOperationsError) as c:
            file_uninstall_handler.uninstall(_file_entry1)
        self.assertEqual(c.exception.app_entry, _file_entry1)
        self.assertEqual(c.exception.root_directory, r'C:\test\test2')


_archive_entry1 = AppEntry(name='123', version='456', archive_url='http://example.com/1.file',
                           unpack_directory=r'C:\test\test2', archive_format='zip', file_hash='789')


class ArchiveInstallHandlerTestCase(TestCase):

    def test_applicability(self):
        self.assertTrue(archive_install_handler.is_applicable(_archive_entry1))

    def test_applicability2(self):
        self.assertFalse(archive_install_handler.is_applicable(replace(_archive_entry1, archive_url=None)))

    def test_applicability3(self):
        self.assertFalse(archive_install_handler.is_applicable(replace(_archive_entry1, archive_url=None)))

    def test_applicability4(self):
        self.assertFalse(archive_install_handler.is_applicable(replace(_archive_entry1, archive_format=None)))

    def test_creating_download_spec(self):
        spec = archive_install_handler.create_download_spec(_archive_entry1, 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/1.file')
        self.assertEqual(spec.download_path, 'C:\\123\\456\\123_456.zip')
        self.assertEqual(spec.part_path, 'C:\\123\\456\\123_456.zip.part')
        self.assertEqual(spec.hash, '789')
        self.assertEqual(spec.name, '123')
        self.assertEqual(spec.version, '456')

    def test_creating_download_spec2(self):
        spec = archive_install_handler.create_download_spec(replace(_archive_entry1, version=None), 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/1.file')
        self.assertEqual(spec.download_path, 'C:\\123\\123.zip')
        self.assertEqual(spec.part_path, 'C:\\123\\123.zip.part')
        self.assertEqual(spec.hash, '789')
        self.assertEqual(spec.name, '123')

    def test_creating_download_spec3(self):
        spec = archive_install_handler.create_download_spec(replace(_archive_entry1, architecture='x86_64'), 'C:\\')
        self.assertEqual(spec.url, 'http://example.com/1.file')
        self.assertEqual(spec.download_path, 'C:\\123\\456\\x86_64\\123_456_x86_64.zip')
        self.assertEqual(spec.part_path, 'C:\\123\\456\\x86_64\\123_456_x86_64.zip.part')
        self.assertEqual(spec.hash, '789')
        self.assertEqual(spec.name, '123')
        self.assertEqual(spec.version, '456')

    def test_installing(self):
        archive_install_handler.install(_archive_entry1, archive_install_handler.create_download_spec(_archive_entry1,
                                                                                                      'C:\\'))
        install_manager.unpack_archive.assert_called_once_with(r'C:\123\456\123_456.zip', 'zip',
                                                               Path('C:/test/test2'))

    def test_installing2(self):
        e = replace(_archive_entry1, unpack_directory=None)
        with self.assertRaises(InvalidOrMissingAppDataError) as c:
            archive_install_handler.install(e, archive_install_handler.create_download_spec(e, 'C:\\'))
        self.assertEqual(c.exception.app_entry, e)
        self.assertEqual(c.exception.key, 'unpack_directory')
        self.assertEqual(c.exception.value, None)

    def test_installing4(self):
        install_manager.unpack_archive.side_effect = OSError
        with self.assertRaises(InterruptedFileOperationsError) as c:
            archive_install_handler.install(_archive_entry1,
                                            archive_install_handler.create_download_spec(_archive_entry1, 'C:\\'))
        self.assertEqual(c.exception.app_entry, _archive_entry1)
        self.assertEqual(c.exception.root_directory, r'C:\test\test2')


class ArchiveUninstallTestCase(InstallManagerMockResetMixin, TestCase):

    def test_applicability(self):
        self.assertTrue(archive_uninstall_handler.is_applicable(_archive_entry1))

    def test_applicability2(self):
        self.assertFalse(archive_uninstall_handler.is_applicable(replace(_archive_entry1, unpack_directory=None)))

    def test_applicability3(self):
        self.assertTrue(archive_uninstall_handler.is_applicable(replace(_archive_entry1, archive_format=None)))

    def test_is_installed(self):
        archive_uninstall_handler.is_installed(_archive_entry1)
        install_manager.is_file_present.assert_called_once_with(r'C:\test\test2')

    def test_uninstalling(self):
        archive_uninstall_handler.uninstall(_archive_entry1)
        install_manager.remove_directory.assert_called_once_with(Path('C:/test/test2'))

    def test_uninstalling2(self):
        e = replace(_archive_entry1, unpack_directory=None)
        with self.assertRaises(InvalidOrMissingAppDataError) as c:
            archive_uninstall_handler.uninstall(e)
        self.assertEqual(c.exception.app_entry, e)
        self.assertEqual(c.exception.key, 'unpack_directory')
        self.assertEqual(c.exception.value, None)

    def test_uninstalling4(self):
        install_manager.remove_directory.side_effect = OSError
        with self.assertRaises(InterruptedFileOperationsError) as c:
            archive_uninstall_handler.uninstall(_archive_entry1)
        self.assertEqual(c.exception.app_entry, _archive_entry1)
        self.assertEqual(c.exception.root_directory, r'C:\test\test2')


class SoftwareProcessorTestCase(InstallManagerMockResetMixin, DownloadManagerMockResetMixin, ExtraAssertMixin,
                                TestCase):

    def test_processing(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_install_manager_data(False, False, '')
        set_download_manager_data({'C:\\temp\\example2\\0.1b\\example2_0.1b.msi'})
        data = get_app_config_from(load_yaml_from_test_dir('app_list'))
        software_processor.process('C:\\temp', data, fb)
        specs = [exe_install_handler.create_download_spec(data.to_install[0], 'C:\\temp'),
                 msi_install_handler.create_download_spec(data.to_install[1], 'C:\\temp'),
                 file_install_handler.create_download_spec(data.to_install[3], 'C:\\temp'),
                 archive_install_handler.create_download_spec(data.to_install[4], 'C:\\temp')]
        download_manager.filter_present.assert_called_once_with(specs)
        download_manager.download.assert_called_once_with(specs[:1] + specs[2:], fb)
        fb.report_checking_software.assert_called_once()
        install_manager.uninstall_msi.assert_not_called()
        install_manager.copy_file.assert_called_once_with('C:\\temp\\example4\\x86_64\\example4_x86_64.exe',
                                                          Path('C:/temp/example4'), 'example4.exe')
        install_manager.unpack_archive.assert_called_once_with(r'C:\temp\example5\0.6\x86_64\example5_0.6_x86_64.zip',
                                                               'zip', Path('C:/temp/example5/0.6'))
        self.assert_called_exactly_with_no_order(install_manager.run_file,
                                                 call('C:\\temp\\example1\\1.1\\example1_1.1.exe',
                                                      ['/SILENT', '/DIR=C:\\Program Files\\Example1\\1.1']))
        self.assert_called_exactly_with_no_order(install_manager.run_command,
                                                 call('manager', ['list']), call('manager', ['list']),
                                                 call('manager', ['install', 'example3:1.1']))
        self.assert_called_exactly_with_no_order(install_manager.install_msi,
                                                 call('C:\\temp\\example2\\0.1b\\example2_0.1b.msi',
                                                      ['/q', 'INSTALLDIR=C:\\Program Files\\Example2']))
        fb.confirm_operations.assert_called_once_with(specs[:1] + specs[2:], [], list(data.to_install))
        fb.report_removal_started.assert_not_called()
        fb.report_removal_finished.assert_not_called()
        fb.report_installation_started.assert_called_once()
        fb.report_installation_finished.assert_called_once()
        self.assert_called_exactly_with_no_order(fb.report_entry_installation_started,
                                                 *(call(item) for item in data.to_install))
        self.assert_called_exactly_with_no_order(fb.report_entry_installation_finished,
                                                 *(call(item) for item in data.to_install))
        fb.report_entry_removal_started.assert_not_called()
        fb.report_entry_removal_finished.assert_not_called()

    def test_processing2(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_install_manager_data(True, True, 'example3_final:1.0\r\nexample3:1.1')
        set_download_manager_data(set())
        data = get_app_config_from(load_yaml_from_test_dir('app_list'))
        software_processor.process('C:\\temp', data, fb)
        fb.report_checking_software.assert_called_once()
        download_manager.filter_present.assert_called_once_with([])
        download_manager.download.assert_not_called()
        install_manager.install_msi.assert_not_called()
        install_manager.remove_file.assert_called_once_with(Path('C:/temp/example4'), 'example4.exe')
        install_manager.remove_directory.assert_called_once_with(Path('C:/temp/example5/0.5'))
        self.assert_called_exactly_with_no_order(install_manager.run_file,
                                                 call('C:\\Program Files\\Example1\\1.0\\uninstall.exe', None))
        self.assert_called_exactly_with_no_order(install_manager.run_command,
                                                 call('manager', ['list']), call('manager', ['list']),
                                                 call('manager', ['uninstall', 'example3_final:1.0']))
        self.assert_called_exactly_with_no_order(install_manager.uninstall_msi, call('pc1235', ['/q']))
        fb.confirm_operations.assert_called_once_with([], list(data.to_uninstall), [])
        fb.report_installation_started.assert_not_called()
        fb.report_installation_finished.assert_not_called()
        fb.report_removal_started.assert_called_once()
        fb.report_removal_finished.assert_called_once()
        self.assert_called_exactly_with_no_order(fb.report_entry_removal_started,
                                                 *(call(item) for item in data.to_uninstall))
        self.assert_called_exactly_with_no_order(fb.report_entry_removal_finished,
                                                 *(call(item) for item in data.to_uninstall))
        fb.report_entry_installation_started.assert_not_called()
        fb.report_entry_installation_finished.assert_not_called()

    def test_processing3(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_download_manager_data(set())
        data = AppData([])
        software_processor.process('C:\\temp', data, fb)
        fb.report_checking_software.assert_called_once()
        download_manager.download.assert_not_called()
        install_manager.install_msi.assert_not_called()
        install_manager.run_file.assert_not_called()
        install_manager.run_command.assert_not_called()
        install_manager.uninstall_msi.assert_not_called()
        install_manager.unpack_archive.assert_not_called()
        install_manager.remove_directory.assert_not_called()
        install_manager.copy_file.assert_not_called()
        install_manager.remove_file.assert_not_called()
        fb.confirm_operations.assert_not_called()
        fb.report_installation_started.assert_not_called()
        fb.report_installation_finished.assert_not_called()
        fb.report_removal_started.assert_not_called()
        fb.report_removal_finished.assert_not_called()
        fb.report_entry_removal_started.assert_not_called()
        fb.report_entry_removal_finished.assert_not_called()
        fb.report_entry_installation_started.assert_not_called()
        fb.report_entry_installation_finished.assert_not_called()
        fb.report_software_set_no_changes.assert_called_once()

    def test_processing4(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_download_manager_data(set())
        data = AppData([AppEntry(name='test', installer_url='http://example.com/file.exe')])
        with self.assertRaises(UnknownUninstallTypeError) as c:
            software_processor.process('C:\\temp', data, fb)
        self.assertEqual(c.exception.app_entry, data.to_uninstall[0])

    def test_processing5(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_download_manager_data(set())
        data = AppData([AppEntry(name='test', uninstaller_path='C:\\uninst.exe', keep=True)])
        with self.assertRaises(UnknownInstallTypeError) as c:
            software_processor.process('C:\\temp', data, fb)
        self.assertEqual(c.exception.app_entry, data.to_install[0])



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest import TestCase
from unittest.mock import patch, DEFAULT

from thespiae import core


@patch.multiple(core, CLI=DEFAULT, parse_config_file=DEFAULT, handle_software_changes=DEFAULT,
                update_user_path=DEFAULT, autospec=True)
class TestMain(TestCase):

    def test_main(self, CLI, parse_config_file, handle_software_changes, update_user_path):
        core.main()
        params = CLI.return_value.__enter__.return_value.params
        fb = CLI.return_value.__enter__.return_value.feedback
        parse_config_file.assert_called_once_with(params.config_file)
        handle_software_changes.assert_called_once_with(params.data_dir, parse_config_file.return_value, fb)
        update_user_path.assert_called_once_with(parse_config_file.return_value, fb)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import asdict
from os.path import splitext
from pathlib import Path
from shutil import get_unpack_formats
from typing import TYPE_CHECKING
from urllib.parse import urlparse

from .data import DownloadSpec
from .exception import InvalidOrMissingAppDataError, UnknownInstallTypeError, UnknownUninstallTypeError, \
    InterruptedFileOperationsError

if TYPE_CHECKING:
    from typing import Optional, Iterable, Tuple

    from thespiae.conf.data import AppData, AppEntry
    from .protocol import Feedback
    from .system import InstallManager, DownloadManager


def _create_download_spec(url: str, directory: str, app_entry: AppEntry, ext: str = None) -> DownloadSpec:
    name = app_entry.name
    if not name:
        raise InvalidOrMissingAppDataError(app_entry, 'name', name)
    parsed = urlparse(url)
    d_dir = Path(directory) / name
    d_name = name
    version = app_entry.version
    if version:
        d_dir /= version
        d_name += f'_{version}'
    architecture = app_entry.architecture
    if architecture:
        d_dir /= architecture
        d_name += f'_{architecture}'
    if not ext:
        _, ext = splitext(parsed.path)
    d_name += ext
    d_path = (d_dir / d_name).absolute()
    return DownloadSpec(**asdict(app_entry), url=url, hash=app_entry.file_hash, download_path=str(d_path),
                        part_path=str(d_path) + '.part')


class InstallTypeHandler(ABC):

    def __init__(self, install_manager: InstallManager):
        self._instm = install_manager

    @abstractmethod
    def create_download_spec(self, app_entry: AppEntry, download_dir_path: str) -> Optional[DownloadSpec]:
        pass

    @abstractmethod
    def is_applicable(self, app_entry: AppEntry) -> bool:
        pass

    @abstractmethod
    def install(self, app_entry: AppEntry, ds: DownloadSpec) -> None:
        pass


class UninstallTypeHandler(ABC):

    def __init__(self, install_manager: InstallManager):
        self._instm = install_manager

    @abstractmethod
    def is_applicable(self, app_entry: AppEntry) -> bool:
        pass

    @abstractmethod
    def is_installed(self, app_entry: AppEntry) -> bool:
        pass

    @abstractmethod
    def uninstall(self, app_entry: AppEntry) -> None:
        pass


class ExeInstallHandler(InstallTypeHandler):

    def create_download_spec(self, app_entry: AppEntry, download_dir_path: str) -> Optional[DownloadSpec]:
        return _create_download_spec(app_entry.installer_url, download_dir_path, app_entry)

    def is_applicable(self, app_entry: AppEntry) -> bool:
        return bool(app_entry.installer_url)

    def install(self, app_entry: AppEntry, ds: DownloadSpec) -> None:
        self._instm.run_file(ds.download_path, app_entry.install_args)


class ExeUninstallHandler(UninstallTypeHandler):

    def is_applicable(self, app_entry: AppEntry) -> bool:
        return bool(app_entry.uninstaller_path)

    def is_installed(self, app_entry: AppEntry) -> bool:
        return self._instm.is_file_present(app_entry.uninstaller_path)

    def uninstall(self, app_entry: AppEntry) -> None:
        self._instm.run_file(app_entry.uninstaller_path, app_entry.uninstall_args)


class MSIInstallHandler(InstallTypeHandler):

    def create_download_spec(self, app_entry: AppEntry, download_dir_path: str) -> Optional[DownloadSpec]:
        return _create_download_spec(app_entry.package_url, download_dir_path, app_entry)

    def is_applicable(self, app_entry: AppEntry) -> bool:
        return bool(app_entry.package_url)

    def install(self, app_entry: AppEntry, ds: DownloadSpec) -> None:
        self._instm.install_msi(ds.download_path, app_entry.install_args)


class MSIUninstallHandler(UninstallTypeHandler):

    def is_applicable(self, app_entry: AppEntry) -> bool:
        return bool(app_entry.product_code)

    def is_installed(self, app_entry: AppEntry) -> bool:
        return self._instm.is_product_code_present(app_entry.product_code)

    def uninstall(self, app_entry: AppEntry) -> None:
        self._instm.uninstall_msi(app_entry.product_code, app_entry.uninstall_args)


class CommandInstallHandler(InstallTypeHandler):

    def create_download_spec(self, app_entry: AppEntry, download_dir_path: str) -> Optional[DownloadSpec]:
        return None

    def is_applicable(self, app_entry: AppEntry) -> bool:
        return bool(app_entry.command)

    def install(self, app_entry: AppEntry, ds: DownloadSpec) -> None:
        self._instm.run_command(app_entry.command, app_entry.install_args)


class CommandUninstallHandler(UninstallTypeHandler):

    def is_applicable(self, app_entry: AppEntry) -> bool:
        return bool(app_entry.command and app_entry.list_args and app_entry.installed_list_entry)

    def is_installed(self, app_entry: AppEntry) -> bool:
        # TODO: may be affected by cmd encoding issues, might need improvements <AP>
        installed_list = self._instm.run_command(app_entry.command, app_entry.list_args)
        return app_entry.installed_list_entry in installed_list

    def uninstall(self, app_entry: AppEntry) -> None:
        self._instm.run_command(app_entry.command, app_entry.uninstall_args)


def _get_path(app_entry: AppEntry, attribute: str) -> Path:
    path = getattr(app_entry, attribute)
    if path:
        return Path(path).absolute()
    else:
        raise InvalidOrMissingAppDataError(app_entry, attribute, path)


class _FileOperations:

    def __init__(self, app_entry: AppEntry, path: Path):
        self.app_entry = app_entry
        self.path = path

    def __enter__(self) -> Path:
        return self.path

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            raise InterruptedFileOperationsError(self.app_entry, str(self.path)) from exc_val


class FileInstallHandler(InstallTypeHandler):

    def create_download_spec(self, app_entry: AppEntry, download_dir_path: str) -> Optional[DownloadSpec]:
        return _create_download_spec(app_entry.file_url, download_dir_path, app_entry)

    def is_applicable(self, app_entry: AppEntry) -> bool:
        return bool(app_entry.file_url) and bool(app_entry.file_name) and bool(app_entry.file_directory)

    def install(self, app_entry: AppEntry, ds: DownloadSpec) -> None:
        with _FileOperations(app_entry, _get_path(app_entry, 'file_directory')) as path:
            self._instm.copy_file(ds.download_path, path, app_entry.file_name)


class FileUninstallHandler(UninstallTypeHandler):

    def is_applicable(self, app_entry: AppEntry) -> bool:
        return bool(app_entry.file_directory) and bool(app_entry.file_name)

    def is_installed(self, app_entry: AppEntry) -> bool:
        return self._instm.is_file_present(str(_get_path(app_entry, 'file_directory') / app_entry.file_name))

    def uninstall(self, app_entry: AppEntry) -> None:
        with _FileOperations(app_entry, _get_path(app_entry, 'file_directory')) as path:
            self._instm.remove_file(path, app_entry.file_name)


_format_extension = {f_name: f_ext[0] for f_name, f_ext, _ in get_unpack_formats()}


class ArchiveInstallHandler(InstallTypeHandler):

    def create_download_spec(self, app_entry: AppEntry, download_dir_path: str) -> Optional[DownloadSpec]:
        return _create_download_spec(app_entry.archive_url, download_dir_path, app_entry,
                                     _format_extension[app_entry.archive_format])

    def is_applicable(self, app_entry: AppEntry) -> bool:
        return bool(app_entry.archive_url and app_entry.archive_format and app_entry.unpack_directory)

    def install(self, app_entry: AppEntry, ds: DownloadSpec) -> None:
        with _FileOperations(app_entry, _get_path(app_entry, 'unpack_directory')) as path:
            self._instm.unpack_archive(ds.download_path, app_entry.archive_format, path)


class ArchiveUninstallHandler(UninstallTypeHandler):

    def is_applicable(self, app_entry: AppEntry) -> bool:
        return bool(app_entry.unpack_directory)

    def is_installed(self, app_entry: AppEntry) -> bool:
        return self._instm.is_file_present(str(_get_path(app_entry, 'unpack_directory')))

    def uninstall(self, app_entry: AppEntry) -> None:
        with _FileOperations(app_entry, _get_path(app_entry, 'unpack_directory')) as path:
            self._instm.remove_directory(path)


class SoftwareProcessor:

    def __init__(self, download_manager: DownloadManager, install_type_handlers: Iterable[InstallTypeHandler],
                 uninstall_type_handlers: Iterable[UninstallTypeHandler]):
        self._dm = download_manager
        self._hs = install_type_handlers
        self._uhs = uninstall_type_handlers

    def _get_handlers(self, app_entry: AppEntry) -> Tuple[InstallTypeHandler, UninstallTypeHandler]:
        for h in self._hs:
            if h.is_applicable(app_entry):
                ih = h
                break
        else:
            raise UnknownInstallTypeError(app_entry)
        for h in self._uhs:
            if h.is_applicable(app_entry):
                uh = h
                break
        else:
            raise UnknownUninstallTypeError(app_entry)
        return ih, uh

    def process(self, download_dir_path: str, config: AppData, fb: Feedback) -> None:
        to_install = []
        to_install_entries = []
        to_uninstall = []
        to_uninstall_entries = []
        specs = []
        fb.report_checking_software()
        for entry in config.to_install:
            install_handler, uninstall_handler = self._get_handlers(entry)
            if not uninstall_handler.is_installed(entry):
                ds = install_handler.create_download_spec(entry, download_dir_path)
                if ds:
                    specs.append(ds)
                to_install.append((entry, install_handler, ds))
                to_install_entries.append(entry)
        for entry in config.to_uninstall:
            install_handler, uninstall_handler = self._get_handlers(entry)
            if uninstall_handler.is_installed(entry):
                to_uninstall.append((entry, uninstall_handler))
                to_uninstall_entries.append(entry)
        specs = self._dm.filter_present(specs)
        if specs or to_install_entries or to_uninstall_entries:
            fb.confirm_operations(specs, to_uninstall_entries, to_install_entries)
            if specs:
                self._dm.download(specs, fb)
            if to_uninstall:
                fb.report_removal_started()
                for entry, handler in to_uninstall:
                    fb.report_entry_removal_started(entry)
                    handler.uninstall(entry)
                    fb.report_entry_removal_finished(entry)
                fb.report_removal_finished()
            if to_install:
                fb.report_installation_started()
                for entry, handler, ds in to_install:
                    fb.report_entry_installation_started(entry)
                    handler.install(entry, ds)
                    fb.report_entry_installation_finished(entry)
                fb.report_installation_finished()
        else:
            fb.report_software_set_no_changes()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from antlr4 import BailErrorStrategy
from antlr4.error.ErrorListener import ConsoleErrorListener

from .envparse.EnvParser import EnvParser as _EnvParser

if TYPE_CHECKING:
    from typing import TextIO

    from antlr4 import TokenStream


class EnvParser(_EnvParser):

    def __init__(self, input_stream: TokenStream, output: TextIO = sys.stdout, build_parse_trees=True):
        super().__init__(input_stream, output)
        self.removeErrorListener(ConsoleErrorListener.INSTANCE)
        self._errHandler = BailErrorStrategy()
        self.buildParseTrees = build_parse_trees



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from dataclasses import dataclass, field
from gettext import translation
from typing import TYPE_CHECKING

from thespiae.exception import ThespiaeError, generate___str__

if TYPE_CHECKING:
    from typing import Sequence, Collection, Mapping, Any

    from .data import AppEntry, ConfigPath

t = translation('thespiae', fallback=True)
_ = t.gettext


@generate___str__
@dataclass(frozen=True)
class AppEntryMixin:
    app_entry: AppEntry = field(metadata={'format': _('configuration entry for: {0.ref}')})


@generate___str__
@dataclass(frozen=True)
class _ConfigPathsMixin:
    config_paths: Sequence[ConfigPath] = field(metadata={'format': _('corresponding paths: {}')})


@generate___str__
@dataclass(frozen=True)
class _ConfigPathMixin:
    config_path: Sequence[ConfigPath] = field(metadata={'format': _('corresponding path: {}')})


@generate___str__
@dataclass(frozen=True)
class _ExpectedValueTypeMixin:
    expected: type = field(metadata={'format': _('expected value type: {0.__name__}')})
    received: type = field(metadata={'format': _('received value type: {0.__name__}')})


@generate___str__
@dataclass(frozen=True)
class _FieldMixin:
    field_name: str = field(metadata={'format': _('config field name: {}')})
    field_index: str = field(metadata={'format': _('collection index: {}')})


@generate___str__(_('Required field is missing'))
@dataclass(frozen=True)
class AppDataMissingFieldError(_FieldMixin, _ConfigPathsMixin, ThespiaeError):
    pass


@generate___str__(_('Unexpected config value'))
@dataclass(frozen=True)
class AppDataFieldValueTypeError(_ExpectedValueTypeMixin, _FieldMixin, _ConfigPathsMixin, ThespiaeError):
    pass


@generate___str__(_('Circular field reference'))
@dataclass(frozen=True)
class AppDataCircularReferenceError(_FieldMixin, _ConfigPathsMixin, ThespiaeError):
    pass


@generate___str__(_('Invalid configuration field value'))
@dataclass(frozen=True)
class ConfigElementTypeError(_ExpectedValueTypeMixin, _ConfigPathMixin, ThespiaeError):
    pass


@generate___str__(_('Excessive configuration attribute found'))
@dataclass(frozen=True)
class ConfigExcessiveAttributeError(_ConfigPathMixin, ThespiaeError):
    pass


@generate___str__(_('Required config attributes not found'))
@dataclass(frozen=True)
class ConfigRequiredAttributesNotFoundError(_ConfigPathMixin, ThespiaeError):
    attributes: Collection[str] = field(metadata={'format': _('missing field names: {}')})


@generate___str__(_('Another configuration entry with the same identity has been found'))
@dataclass(frozen=True)
class ConfigDuplicatedEntryIdentityError(_ConfigPathsMixin, ThespiaeError):
    another_paths: Sequence[ConfigPath] = field(metadata={'format': _('another paths: {}')})
    identity_values: Mapping[str, Any] = field(metadata={'format': _('identity field values: {}')})


@generate___str__(_('Unable to complete config branches with required identity attributes'))
@dataclass(frozen=True)
class ConfigIncompleteBranchesError(_ConfigPathsMixin, ThespiaeError):
    attributes: Collection[str] = field(metadata={'format': _('missing field names: {}')})


class _CircularReferenceError(ThespiaeError):
    pass



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import builtins
from unittest import TestCase
from unittest.mock import patch, DEFAULT

from thespiae.conf import system as ts
from .singleton import real_app_data_reader as app_data_reader


class AppDataReaderTestCase(TestCase):

    @patch.multiple(builtins, open=DEFAULT)
    @patch.multiple(ts, safe_load=DEFAULT)
    def test_reading_config_file(self, safe_load, open):
        open.return_value.__enter__.return_value = open.return_value
        app_data_reader.read('test_path')
        open.assert_called_once_with('test_path', mode='tr')
        safe_load.assert_called_once_with(open.return_value)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .core import ExeInstallHandler, MSIInstallHandler, CommandInstallHandler, SoftwareProcessor, ExeUninstallHandler, \
    MSIUninstallHandler, CommandUninstallHandler, FileInstallHandler, FileUninstallHandler, ArchiveInstallHandler, \
    ArchiveUninstallHandler
from .system import InstallManager, DownloadManager

install_manager = InstallManager()
download_manager = DownloadManager()
software_processor = SoftwareProcessor(download_manager, [ExeInstallHandler(install_manager),
                                                          MSIInstallHandler(install_manager),
                                                          CommandInstallHandler(install_manager),
                                                          FileInstallHandler(install_manager),
                                                          ArchiveInstallHandler(install_manager)],
                                       [ExeUninstallHandler(install_manager),
                                        MSIUninstallHandler(install_manager),
                                        CommandUninstallHandler(install_manager),
                                        FileUninstallHandler(install_manager),
                                        ArchiveUninstallHandler(install_manager)])



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from os.path import join, dirname, isabs

from yaml import safe_load

from .singleton import app_data_reader


def get_test_file_path(file_name: str):
    if isabs(file_name):
        return file_name
    else:
        return join(dirname(__file__), file_name if file_name.endswith('.yml') else file_name + '.yml')


def load_yaml_from_test_dir(file_name: str):
    with open(get_test_file_path(file_name), mode='rb') as f:
        return safe_load(f)


def set_app_data_reader_data(path: str):
    app_data_reader.read.side_effect = lambda _: load_yaml_from_test_dir(path)


class AppDataReaderResetMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        app_data_reader.reset_mock(side_effect=True)
        app_data_reader.read.reset_mock(side_effect=True)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from asyncio import gather, create_task, get_running_loop, run as aio_run, set_event_loop_policy, \
    DefaultEventLoopPolicy, WindowsSelectorEventLoopPolicy
from concurrent.futures import ThreadPoolExecutor
from hashlib import sha256
from io import SEEK_END
from os import rename, makedirs, rmdir, listdir, remove, removedirs
from os.path import exists, dirname
from re import fullmatch, ASCII
from shutil import copyfile, rmtree, unpack_archive
from subprocess import run as subprocess_run, CalledProcessError
import sys
from typing import TYPE_CHECKING
from winreg import OpenKeyEx, HKEY_LOCAL_MACHINE, HKEY_CURRENT_USER, KEY_WOW64_32KEY, KEY_READ, KEY_WOW64_64KEY

from aiohttp import ClientSession, ClientError
from aiohttp.web import HTTPOk, HTTPPartialContent, HTTPRequestRangeNotSatisfiable

from .data import ContentRangeData
from .exception import DownloadError, HashError, UnsupportedResponseError

if TYPE_CHECKING:
    from typing import MutableMapping, Mapping, Sequence, List, Optional, BinaryIO
    from pathlib import Path

    from aiohttp import ClientResponse

    from .data import DownloadSpec
    from .protocol import Feedback

# TODO: remove for 4.x+ version of aiohttp (https://github.com/aio-libs/aiohttp/issues/4324) <AP>
if sys.platform == 'win32':
    set_event_loop_policy(WindowsSelectorEventLoopPolicy())


def _test_key_present_view(root: int, path: str, key: str, view: int) -> bool:
    try:
        with OpenKeyEx(root, f'{path}\\{key}', 0, KEY_READ | view):
            return True
    except OSError as e:
        if e.winerror == 2:
            return False
        else:
            raise


def _test_key_present(root: int, path: str, key: str) -> bool:
    return _test_key_present_view(root, path, key, KEY_WOW64_64KEY) or \
           _test_key_present_view(root, path, key, KEY_WOW64_32KEY)


class InstallManager:

    @staticmethod
    def is_file_present(path: str) -> bool:
        return exists(path)

    @staticmethod
    def is_product_code_present(code: str) -> bool:
        # TODO: 32-bit installer compatibility <AP>
        return _test_key_present(HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall', code) or \
               _test_key_present(HKEY_CURRENT_USER, r'SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall', code)

    @staticmethod
    def run_file(path: str, args: List[str]) -> None:
        subprocess_run([path] + (args if args else []), check=True, shell=True)

    def install_msi(self, package_path: str, args: List[str]) -> None:
        self.run_command('msiexec', ['/I', package_path] + (args if args else []))

    def uninstall_msi(self, product_code: str, args: List[str]) -> None:
        self.run_command('msiexec', ['/x', product_code] + (args if args else []))

    @staticmethod
    def run_command(command: str, args: List[str]) -> str:
        try:
            return subprocess_run([command] + (args if args else []), shell=True, capture_output=True, text=True,
                                  check=True).stdout
        except CalledProcessError as e:
            if e.returncode != 3010:
                raise

    @staticmethod
    def copy_file(src: str, path: Path, name: str):
        makedirs(path, exist_ok=True)
        copyfile(src, path / name)

    @staticmethod
    def _remove_directories(path: Sequence[Path]):
        # TODO: investigate options to raise exception if directory is in use <AP>
        for p in reversed(path):
            p_str = str(p)
            if not listdir(p_str):
                rmdir(p_str)

    @staticmethod
    def remove_directory(path: Path):
        rmtree(path)
        removedirs(path.parent)

    @staticmethod
    def remove_file(path: Path, name: str):
        remove(path / name)
        removedirs(path)

    @staticmethod
    def unpack_archive(archive_path: str, archive_format: str, directory_path: Path):
        makedirs(directory_path, exist_ok=True)
        unpack_archive(archive_path, directory_path, archive_format)


def _prepare_download(part_path: str, fos: MutableMapping[str, BinaryIO]) -> int:
    if part_path in fos:
        raise RuntimeError(f'{part_path} is already associated with another download')
    if exists(part_path):
        fo = open(part_path, 'r+b')
        fo.seek(0, SEEK_END)
        ln = fo.tell()
    else:
        makedirs(dirname(part_path), exist_ok=True)
        fo = open(part_path, 'w+b')
        ln = 0
    fos[part_path] = fo
    return ln


def _write_data(part_path: str, data: bytes, fos: Mapping[str, BinaryIO]) -> None:
    fo = fos[part_path]
    fo.write(data)


def _truncate_file(part_path: str, fos: Mapping[str, BinaryIO]) -> None:
    fo = fos[part_path]
    fo.seek(0)
    fo.truncate(0)


def _finalize_download(part_path: str, download_path: str, fos: MutableMapping[str, BinaryIO]) -> None:
    fo = fos.pop(part_path)
    fo.close()
    rename(part_path, download_path)


def _close_fos(fos) -> None:
    for fo in fos.values():
        fo.close()


def parse_content_range(val: str) -> Optional[ContentRangeData]:
    mo = fullmatch(r'bytes (?:\*|(?P<start>\d+)-(?P<end>\d+))/(?:\*|(?P<total>\d+))', val, ASCII)
    if mo:
        start, end, total = mo.group('start', 'end', 'total')
        return ContentRangeData(int(start) if start else None, int(end) if end else None,
                                int(total) if total else None)
    else:
        return None


class DownloadManager:

    def __init__(self):
        self._ex = ThreadPoolExecutor(max_workers=1)
        self._cs = 1000 * 1000

    def _get_hash(self, path) -> str:
        h = sha256()
        cs = self._cs
        with open(path, 'rb') as fo:
            data = fo.read(cs)
            while len(data) > 0:
                h.update(data)
                data = fo.read(cs)
        return h.hexdigest()

    async def _fetch_and_write(self, resp: ClientResponse, spec: DownloadSpec, fos: MutableMapping[str, BinaryIO],
                               fb: Feedback) -> None:
        loop = get_running_loop()
        cs = self._cs
        ex = self._ex
        path = spec.part_path
        while True:
            data = await resp.content.read(cs)
            if data:
                await loop.run_in_executor(ex, _write_data, path, data, fos)
                fb.report_entry_download_progress(spec, len(data))
            else:
                await loop.run_in_executor(ex, _finalize_download, path, spec.download_path, fos)
                fb.report_entry_download_finished(spec)
                break

    async def _download_file(self, session: ClientSession, spec: DownloadSpec, fos: MutableMapping[str, BinaryIO],
                             fb: Feedback) -> None:
        loop = get_running_loop()
        fb.report_entry_download_initiated(spec)
        try:
            ln = await loop.run_in_executor(self._ex, _prepare_download, spec.part_path, fos)
            headers = {}
            if ln:
                headers['Range'] = f'bytes={ln}-'
            async with session.get(spec.url, headers=headers) as resp:
                loop = get_running_loop()
                if resp.status == HTTPOk.status_code:
                    if ln:
                        await loop.run_in_executor(self._ex, _truncate_file, spec.part_path, fos)
                        ln = 0
                    fb.report_entry_download_started(spec, ln, resp.content_length)
                    await self._fetch_and_write(resp, spec, fos, fb)
                elif resp.status == HTTPPartialContent.status_code:
                    fb.report_entry_download_started(spec, ln,
                                                     resp.content_length + ln if resp.content_length else None)
                    await self._fetch_and_write(resp, spec, fos, fb)
                elif resp.status == HTTPRequestRangeNotSatisfiable.status_code:
                    crd = resp.headers.get('Content-Range')
                    if crd:
                        parsed_crd = parse_content_range(crd)
                        if parsed_crd and parsed_crd.total == ln:
                            await loop.run_in_executor(self._ex, _finalize_download, spec.part_path,
                                                       spec.download_path, fos)
                            fb.report_entry_download_started(spec, ln, ln)
                            fb.report_entry_download_finished(spec)
                        else:
                            raise UnsupportedResponseError(spec, resp.status, resp.headers)
                    else:
                        raise UnsupportedResponseError(spec, resp.status, resp.headers)
                else:
                    raise UnsupportedResponseError(spec, resp.status, resp.headers)
        except (OSError, ClientError) as e:
            raise DownloadError(spec) from e

    async def _download(self, specs: Sequence[DownloadSpec], fb: Feedback) -> None:
        async with ClientSession() as session:
            fos = {}
            tasks = [create_task(self._download_file(session, s, fos, fb)) for s in specs]
            try:
                await gather(*tasks)
            finally:
                to_wait_for = []
                for task in tasks:
                    if not task.done():
                        task.cancel()
                        to_wait_for.append(task)
                await gather(*to_wait_for, return_exceptions=True)
                loop = get_running_loop()
                try:
                    await loop.run_in_executor(self._ex, _close_fos, fos)
                except OSError:
                    # TODO: logging <AP>
                    pass

    def _ensure_hash(self, spec: DownloadSpec) -> None:
        if spec.hash:
            actual = self._get_hash(spec.download_path)
            if actual != spec.hash:
                raise HashError(spec, actual)

    def filter_present(self, specs: Sequence[DownloadSpec]) -> List[DownloadSpec]:
        a = []
        for spec in specs:
            if exists(spec.download_path):
                self._ensure_hash(spec)
            else:
                a.append(spec)
        return a

    def download(self, specs: Sequence[DownloadSpec], fb: Feedback) -> None:
        async def _helper(std):
            download_task = create_task(self._download(std, fb))
            await download_task

        if specs:
            fb.report_download_started()
            aio_run(_helper(specs))
        for s in specs:
            self._ensure_hash(s)
        fb.report_download_finished()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from dataclasses import dataclass, field
from typing import Sequence, Mapping
from unittest import TestCase

from thespiae.exception import generate___str__, META_FORMAT


class ExceptionFormattingTestCase(TestCase):

    def test_formatting(self):
        @generate___str__('test message')
        @dataclass(frozen=True)
        class TestError:
            pass

        te = TestError()
        self.assertEqual(str(te), 'test message')

    def test_formatting2(self):
        @generate___str__
        @dataclass(frozen=True)
        class TestError:
            field1: int = field(metadata={META_FORMAT: 'field #1: {}'})
            field2: int = field(metadata={META_FORMAT: 'field #2: {}'})

        te = TestError('a', 1)
        self.assertEqual(str(te), 'field #1: a, field #2: 1')

    def test_formatting3(self):
        @generate___str__('test message')
        @dataclass(frozen=True)
        class TestError:
            field1: int = field(metadata={META_FORMAT: 'field #1: {}'})
            field2: int = field(metadata={META_FORMAT: 'field #2: {}'})

        te = TestError('a', 1)
        self.assertEqual(str(te), 'test message; field #1: a, field #2: 1')

    def test_formatting4(self):
        @generate___str__('test message')
        @dataclass(frozen=True)
        class TestError:
            field1: str = field(metadata={META_FORMAT: 'field #1: {}'})
            field2: int = field(metadata={META_FORMAT: 'field #2: {}'})

        @generate___str__('test message 2')
        @dataclass(frozen=True)
        class TestError2(TestError):
            field3: Sequence[str] = field(metadata={META_FORMAT: 'field #3: {}'})
            field4: Mapping[str, int] = field(metadata={META_FORMAT: 'field #4: {}'})

        te = TestError2('a', 1, ['c', 'd'], {'e': 2})
        self.assertEqual(str(te), 'test message 2; field #1: a, field #2: 1, field #3: c, d, field #4: e:2')



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest.mock import NonCallableMock

from thespiae.install.core import ExeInstallHandler, MSIInstallHandler, CommandInstallHandler, SoftwareProcessor, \
    ExeUninstallHandler, MSIUninstallHandler, CommandUninstallHandler, FileInstallHandler, FileUninstallHandler, \
    ArchiveInstallHandler, ArchiveUninstallHandler
from thespiae.install.system import InstallManager, DownloadManager

install_manager = NonCallableMock(spec_set=InstallManager)
download_manager = NonCallableMock(spec_set=DownloadManager)
exe_install_handler = ExeInstallHandler(install_manager)
msi_install_handler = MSIInstallHandler(install_manager)
command_install_handler = CommandInstallHandler(install_manager)
file_install_handler = FileInstallHandler(install_manager)
exe_uninstall_handler = ExeUninstallHandler(install_manager)
msi_uninstall_handler = MSIUninstallHandler(install_manager)
command_uninstall_handler = CommandUninstallHandler(install_manager)
file_uninstall_handler = FileUninstallHandler(install_manager)
archive_install_handler = ArchiveInstallHandler(install_manager)
archive_uninstall_handler = ArchiveUninstallHandler(install_manager)
software_processor = SoftwareProcessor(download_manager, [exe_install_handler,
                                                          msi_install_handler,
                                                          command_install_handler,
                                                          file_install_handler,
                                                          archive_install_handler],
                                       [exe_uninstall_handler,
                                        msi_uninstall_handler,
                                        command_uninstall_handler,
                                        file_uninstall_handler,
                                        archive_uninstall_handler])

real_download_manager = DownloadManager()
real_install_manager = InstallManager()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest import TestCase

from thespiae.conf.core import create_app_entry, get_app_config_from
from thespiae.conf.data import AppEntry, ConfigPath
from thespiae.conf.exception import AppDataMissingFieldError, \
    AppDataFieldValueTypeError, AppDataCircularReferenceError, ConfigElementTypeError, \
    ConfigDuplicatedEntryIdentityError, ConfigIncompleteBranchesError, ConfigExcessiveAttributeError, \
    ConfigRequiredAttributesNotFoundError
from .helper import load_yaml_from_test_dir, set_app_data_reader_data, AppDataReaderResetMixin
from .singleton import config_processor, app_data_reader


class AppEntryCreationTest(TestCase):

    def test_app_entry_creation(self):
        e = create_app_entry([{'name': 'example', 'version': '1.1'}], [])
        self.assertEqual(e.name, 'example')
        self.assertEqual(e.version, '1.1')

    def test_no_name(self):
        with self.assertRaises(AppDataMissingFieldError) as c:
            create_app_entry([{'version': '1.1'}], [ConfigPath('$[1]'), ConfigPath('$[1].versions[2]')])
        self.assertEqual(c.exception.field_name, 'name')
        self.assertSequenceEqual(c.exception.config_paths, [ConfigPath('$[1]'), ConfigPath('$[1].versions[2]')])

    def test_int_instead_str(self):
        with self.assertRaises(AppDataFieldValueTypeError) as c:
            create_app_entry([{'name': 1}], [ConfigPath('$[2]')])
        self.assertEqual(c.exception.expected, str)
        self.assertEqual(c.exception.received, int)
        self.assertEqual(c.exception.field_name, 'name')
        self.assertSequenceEqual(c.exception.config_paths, [ConfigPath('$[2]')])

    def test_int_list_instead_str_list(self):
        with self.assertRaises(AppDataFieldValueTypeError) as c:
            create_app_entry([{'name': 'example', 'version': '1.1', 'install_args': [1, 2, 3]}], [ConfigPath('$[3]')])
        self.assertEqual(c.exception.expected, str)
        self.assertEqual(c.exception.received, int)
        self.assertEqual(c.exception.field_name, 'install_args')
        self.assertSequenceEqual(c.exception.config_paths, [ConfigPath('$[3]')])

    def test_str_instead_list(self):
        with self.assertRaises(AppDataFieldValueTypeError) as c:
            create_app_entry([{'name': 'example', 'version': '1.1', 'uninstall_args': 'abc'}],
                             [ConfigPath('$[9]'), ConfigPath('$[9].versions[5]')])
        self.assertEqual(c.exception.expected, list)
        self.assertEqual(c.exception.received, str)
        self.assertEqual(c.exception.field_name, 'uninstall_args')
        self.assertSequenceEqual(c.exception.config_paths, [ConfigPath('$[9]'), ConfigPath('$[9].versions[5]')])

    def test_refinement(self):
        e = create_app_entry([{'name': 'example', 'version': '1.1'}, {'install_args': ['a', 'b', 'c']}], [])
        self.assertSequenceEqual(e.install_args, ['a', 'b', 'c'])

    def test_refinement2(self):
        e = create_app_entry([{'name': 'example', 'version': '1.1', 'install_args': ['a', 'b', 'c']},
                              {'install_args': ['a', 'b']}], [])
        self.assertSequenceEqual(e.install_args, ['a', 'b'])

    def test_refinement3(self):
        e = create_app_entry([{'name': 'example', 'version': '1.1', 'command': '$a', 'b': 'test'}, {'a': '$b'}], [])
        self.assertEqual(e.command, 'test')

    def test_refinement4(self):
        e = create_app_entry([{'name': 'example', 'version': '1.1', 'command': 'te${a}t'}, {'a': 's'}], [])
        self.assertEqual(e.command, 'test')

    def test_list_refinement(self):
        e = create_app_entry([{'name': 'example', 'version': '1.1', 'install_args': ['a${b}c', 'de$f']},
                              {'b': 0, 'f': 1}], [])
        self.assertSequenceEqual(e.install_args, ['a0c', 'de1'])

    def test_circular_reference(self):
        with self.assertRaises(AppDataCircularReferenceError) as c:
            create_app_entry([{'name': 'example', 'version': '1.1', 'command': '$command'}], [ConfigPath('$[1]')])
        self.assertEqual(c.exception.field_name, 'command')
        self.assertSequenceEqual(c.exception.config_paths, [ConfigPath('$[1]')])

    def test_circular_reference2(self):
        with self.assertRaises(AppDataCircularReferenceError) as c:
            create_app_entry([{'name': 'example', 'version': '1.1', 'command': '$a', 'a': '$command'}],
                             [ConfigPath('$[1]')])
        self.assertEqual(c.exception.field_name, 'command')
        self.assertSequenceEqual(c.exception.config_paths, [ConfigPath('$[1]')])

    def test_circular_reference3(self):
        with self.assertRaises(AppDataCircularReferenceError) as c:
            create_app_entry([{'name': 'example', 'version': '1.1', 'command': '$a'}, {'a': '$command'}],
                             [ConfigPath('$[1]'), ConfigPath('$[1].versions[2]')])
        self.assertEqual(c.exception.field_name, 'command')
        self.assertSequenceEqual(c.exception.config_paths, [ConfigPath('$[1]'), ConfigPath('$[1].versions[2]')])

    def test_refinement_priority(self):
        e = create_app_entry([{'name': 'example', 'version': '1.1', 'command': '$a', 'a': 'test'}, {'a': 'test2'}], [])
        self.assertEqual(e.command, 'test2')


class AppDataCreationTest(TestCase):

    def test_app_list_processing(self):
        app_conf = get_app_config_from(load_yaml_from_test_dir('app_list'))
        self.assertEqual(len(app_conf.to_install), 5)
        self.assertEqual(len(app_conf.to_uninstall), 5)

    def test_app_list_processing2(self):
        with self.assertRaises(ConfigDuplicatedEntryIdentityError) as c:
            get_app_config_from(load_yaml_from_test_dir('app_list_duplicated_name_version'))
        self.assertSequenceEqual(c.exception.config_paths, [ConfigPath('$[0]'), ConfigPath('$[0].versions[0]')])
        self.assertSequenceEqual(c.exception.another_paths, [ConfigPath('$[1]'), ConfigPath('$[1].versions[0]')])
        self.assertDictEqual(c.exception.identity_values, {'architecture': None, 'name': 'test', 'version': '1.0'})

    def test_app_list_processing3(self):
        ac = get_app_config_from(load_yaml_from_test_dir('app_list_no_versions'))
        self.assertSequenceEqual(ac.to_uninstall, [AppEntry(name='test2', version='1.0'),
                                                   AppEntry(name='test', version='1.0')])

    def test_config_structure(self):
        with self.assertRaises(ConfigElementTypeError) as c:
            get_app_config_from(load_yaml_from_test_dir('app_list_invalid_root'))
        self.assertEqual(c.exception.expected, list)
        self.assertEqual(c.exception.received, str)
        self.assertEqual(c.exception.config_path, ConfigPath('$'))

    def test_config_structure2(self):
        with self.assertRaises(ConfigElementTypeError) as c:
            get_app_config_from(load_yaml_from_test_dir('app_list_invalid_root_entry'))
        self.assertEqual(c.exception.expected, dict)
        self.assertEqual(c.exception.received, str)
        self.assertEqual(c.exception.config_path, ConfigPath('$[1]'))

    def test_config_structure3(self):
        with self.assertRaises(ConfigElementTypeError) as c:
            get_app_config_from(load_yaml_from_test_dir('app_list_invalid_versions'))
        self.assertEqual(c.exception.expected, list)
        self.assertEqual(c.exception.received, str)
        self.assertEqual(c.exception.config_path, ConfigPath('$[1].versions'))

    def test_config_structure4(self):
        with self.assertRaises(ConfigElementTypeError) as c:
            get_app_config_from(load_yaml_from_test_dir('app_list_invalid_version_entry'))
        self.assertEqual(c.exception.expected, dict)
        self.assertEqual(c.exception.received, str)
        self.assertEqual(c.exception.config_path, ConfigPath('$[1].versions[1]'))


class ConfigProcessorTest(AppDataReaderResetMixin, TestCase):

    def test_config_processing(self):
        set_app_data_reader_data('app_list')
        data = config_processor.process_config('config.yml')
        app_data_reader.read.assert_called_once_with('config.yml')
        to_install = [AppEntry(name='example1', installer_url='http://example1.com/test_1.1.exe', file_hash='1234',
                               uninstaller_path='C:\\Program Files\\Example1\\1.1\\uninstall.exe',
                               install_args=['/SILENT', '/DIR=C:\\Program Files\\Example1\\1.1'],
                               path_entries=['C:\\Program Files\\Example1\\1.1\\bin',
                                             'C:\\Program Files\\Example1\\1.1\\bin2'], version='1.1', keep=True),
                      AppEntry(name='example2', package_url='http://example.com/example2_0.1b.msi', file_hash='5678',
                               product_code='pc1234', install_args=['/q', 'INSTALLDIR=C:\\Program Files\\Example2'],
                               path_entries=['C:\\Program Files\\Example2'], version='0.1b',
                               keep=True),
                      AppEntry(name='example3', command='manager', install_args=['install', 'example3:1.1'],
                               uninstall_args=['uninstall', 'example3:1.1'],
                               path_entries=['C:\\apps\\example3\\1.1\\bin'], list_args=['list'],
                               installed_list_entry='example3:1.1', version='1.1', keep=True),
                      AppEntry(name='example4', architecture='x86_64', file_directory=r'C:\temp\example4',
                               file_url='http://example.com/x86_64/file.exe', file_name='example4.exe',
                               file_hash='456', keep=True),
                      AppEntry(name='example5', architecture='x86_64', unpack_directory=r'C:\temp\example5\0.6',
                               archive_url='http://example.com/x86_64/file-0.6.zip', archive_format='zip',
                               file_hash='123', keep=True, version='0.6')]
        to_uninstall = [AppEntry(name='example1', installer_url='http://example1.com/test_1.0.exe', file_hash='1235',
                                 uninstaller_path='C:\\Program Files\\Example1\\1.0\\uninstall.exe',
                                 install_args=['/SILENT', '/DIR=C:\\Program Files\\Example1\\1.0'],
                                 path_entries=['C:\\Program Files\\Example1\\1.0\\bin',
                                               'C:\\Program Files\\Example1\\1.0\\bin2'], version='1.0'),
                        AppEntry(name='example3', command='manager', install_args=['install', 'example3_final:1.0'],
                                 uninstall_args=['uninstall', 'example3_final:1.0'],
                                 path_entries=['C:\\apps\\example3\\1.0\\bin'], list_args=['list'],
                                 installed_list_entry='example3_final:1.0', version='1.0'),
                        AppEntry(name='example2', package_url='http://example.com/example2_0.1a.msi', file_hash='3579',
                                 product_code='pc1235', install_args=['/q', 'INSTALLDIR=C:\\Program Files\\Ex2'],
                                 uninstall_args=['/q'], path_entries=['C:\\Program Files\\Ex2'], version='0.1a'),
                        AppEntry(name='example4', architecture='x86', file_directory=r'C:\temp\example4',
                                 file_url='http://example.com/x86/file.exe', file_name='example4.exe',
                                 file_hash='123'),
                        AppEntry(name='example5', architecture='x86_64',
                                 unpack_directory=r'C:\temp\example5\0.5',
                                 archive_url='http://example.com/x86_64/file-0.5.zip', archive_format='zip',
                                 file_hash='456', version='0.5')]
        self.assertSequenceEqual(data.to_install, to_install)
        self.assertSequenceEqual(data.to_uninstall, to_uninstall)

    def test_config_processing2(self):
        set_app_data_reader_data('app_entry_name_only')
        data = config_processor.process_config('config.yml')
        self.assertEqual(len(data.to_install), 0)
        self.assertEqual(len(data.to_uninstall), 1)
        self.assertEqual(data.to_uninstall[0].name, 'example1')
        self.assertEqual(data.to_uninstall[0].file_hash, '123')


class NewConfigProcessorTest(TestCase):

    def test_config_data_processing(self):
        data = load_yaml_from_test_dir('advanced_config_structure')
        items = get_app_config_from(data)
        to_install = [AppEntry(name='test', version='0.3', architecture='x86_64', keep=True, file_hash='5'),
                      AppEntry(name='test', version='0.2', architecture='x86_64', keep=True, file_hash='6'),
                      AppEntry(name='test2', version='0.3', architecture='x86_64', keep=True, file_hash='5'),
                      AppEntry(name='test2', version='0.1', architecture='x86', keep=True, file_hash='7')]
        to_uninstall = [AppEntry(name='test', version='999.1', architecture='x86_128', file_hash='4'),
                        AppEntry(name='test', version='0.1', architecture='x86', file_hash='7'),
                        AppEntry(name='test2', version='999.1', architecture='x86_128', file_hash='4'),
                        AppEntry(name='test2', version='0.2', architecture='x86_64', file_hash='6')]
        self.assertSequenceEqual(items.to_install, to_install)
        self.assertSequenceEqual(items.to_uninstall, to_uninstall)

    def test_no_name(self):
        with self.assertRaises(ConfigRequiredAttributesNotFoundError) as c:
            get_app_config_from(load_yaml_from_test_dir('advanced_config_no_name'))
        self.assertEqual(c.exception.config_path, ConfigPath('$[1]'))
        self.assertSetEqual(c.exception.attributes, {'name'})

    def test_excessive_attribute(self):
        with self.assertRaises(ConfigExcessiveAttributeError) as c:
            get_app_config_from(load_yaml_from_test_dir('advanced_config_excessive_attribute.yml'))
        self.assertEqual(c.exception.config_path, ConfigPath('$[0].versions[0].versions'))

    def test_excessive_attribute2(self):
        with self.assertRaises(ConfigExcessiveAttributeError) as c:
            get_app_config_from(load_yaml_from_test_dir('advanced_config_excessive_attribute2.yml'))
        self.assertEqual(c.exception.config_path, ConfigPath('$[0].architectures[0].version'))

    def test_excessive_attribute3(self):
        with self.assertRaises(ConfigExcessiveAttributeError) as c:
            get_app_config_from(load_yaml_from_test_dir('advanced_config_excessive_attribute3.yml'))
        self.assertEqual(c.exception.config_path, ConfigPath('$[0].versions'))

    def test_incomplete_branches(self):
        with self.assertRaises(ConfigIncompleteBranchesError) as c:
            get_app_config_from(load_yaml_from_test_dir('advanced_config_incomplete_branches.yml'))
        self.assertSequenceEqual(c.exception.config_paths, [ConfigPath('$[0].architectures[0]')])
        self.assertSetEqual(c.exception.attributes, {'version'})



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from collections import defaultdict
from dataclasses import fields, MISSING, asdict
from string import Template
from typing import TYPE_CHECKING, Mapping, List, Iterable, MutableMapping

from .data import AppEntry, AppData, ConfigPath, AppEntryRef, _FORK_ATTR, _GROUP_ATTR
from .exception import _CircularReferenceError, AppDataMissingFieldError, \
    AppDataFieldValueTypeError, AppDataCircularReferenceError, ConfigElementTypeError, ConfigExcessiveAttributeError, \
    ConfigRequiredAttributesNotFoundError, ConfigDuplicatedEntryIdentityError, ConfigIncompleteBranchesError

if TYPE_CHECKING:
    from typing import Sequence, FrozenSet, Set, MutableSet, Any, Optional, Tuple, Iterator

    from .system import AppDataReader

_fork_eliminations = {field.name: field.metadata[_FORK_ATTR] for field in fields(AppEntryRef)}
_group_by = frozenset((field.name for field in fields(AppEntryRef) if field.metadata.get(_GROUP_ATTR)))
_fork_set = frozenset(_fork_eliminations.values())


class _LazyExpandedDict(Mapping):

    def __iter__(self):
        return iter(self.raw)

    def __len__(self):
        return len(self.raw)

    def __new__(cls, raw: Mapping, refined: Mapping = None):
        if raw is not None:
            return super().__new__(cls, raw, refined)
        else:
            return None

    def __init__(self, raw: Mapping, refined: Mapping = None):
        self.raw = raw
        self.refined = refined
        self._final = {}
        self._breaker = object()

    def _format_helper(self, data):
        if isinstance(data, str):
            if self.refined is not None:
                template = Template(data)
                data = template.safe_substitute(self.refined)
            while True:
                prev = data
                template = Template(prev)
                data = template.safe_substitute(self)
                if prev == data:
                    return data
        elif isinstance(data, list):
            return [self._format_helper(e) for e in data]
        else:
            return data

    def __getitem__(self, key):
        data = self._final.get(key)
        if data is self._breaker:
            raise _CircularReferenceError
        if data is not None:
            return data
        raw_data = self.raw.get(key)
        if raw_data is None:
            raise KeyError
        self._final[key] = self._breaker
        final_data = self._format_helper(raw_data)
        self._final[key] = final_data
        return final_data


# TODO: support type unions <AP>
def _validate_types(v, t, f_name: str, paths: Sequence[ConfigPath], position: int = None) -> None:
    ot = getattr(t, '__origin__', None)
    if ot is None:
        if not isinstance(v, t):
            raise AppDataFieldValueTypeError(paths, f_name, position, t, type(v))
    else:
        if issubclass(ot, Iterable):
            if isinstance(v, ot):
                if len(t.__args__) == 1:
                    for p, sv in enumerate(v):
                        _validate_types(sv, t.__args__[0], f_name, paths, p)
                else:
                    raise RuntimeError('Unsupported generic type', t)
            else:
                raise AppDataFieldValueTypeError(paths, f_name, None, ot, type(v))
        else:
            raise RuntimeError('Unsupported generic type', t)


def create_app_entry(namespaces: Iterable[Mapping], paths: Sequence[ConfigPath]) -> AppEntry:
    lazy_dicts = []
    ld = None
    for d in namespaces:
        ld = _LazyExpandedDict(d, ld)
        lazy_dicts.append(ld)
    if len(lazy_dicts) > 1:
        lazy_dicts[0].refined = lazy_dicts[-1]
        lazy_dicts.reverse()
    data = {}
    for field in fields(AppEntry):
        value = None
        try:
            for d in lazy_dicts:
                value = d.get(field.name)
                if value is not None:
                    break
        except _CircularReferenceError:
            raise AppDataCircularReferenceError(paths, field.name, None)
        if value is not None:
            try:
                _validate_types(value, eval(field.type), field.name, paths, None)
            except NameError:
                raise RuntimeError(f'Unknown type {field.type}, probably lacking import')
            data[field.name] = value
        else:
            if field.default is MISSING and field.default_factory is MISSING:
                raise AppDataMissingFieldError(paths, field.name, None)
    return AppEntry(**data)


class _Indexer:

    def __init__(self, start=0):
        self._c = start

    @property
    def current(self) -> int:
        return self._c

    def inc(self) -> None:
        self._c += 1


def _process_config_data(parent_index: Optional[int], path: ConfigPath, branch_fork_set: FrozenSet[str],
                         branch_attrs: FrozenSet[str], config_data: Any, c: _Indexer) -> Iterator[Optional[int],
                                                                                                  int, ConfigPath, Set,
                                                                                                  bool, Mapping]:
    if isinstance(config_data, list):
        for n, config_item in enumerate(config_data):
            current_path = path.index(n)
            if isinstance(config_item, MutableMapping):
                current_branch_attrs = set()
                for key in config_item:
                    if key in branch_attrs:
                        raise ConfigExcessiveAttributeError(current_path.attribute(key))
                    if key in branch_fork_set:
                        raise ConfigExcessiveAttributeError(current_path.attribute(key))
                    if key in _fork_eliminations:
                        if _fork_eliminations[key] in config_item:
                            raise ConfigExcessiveAttributeError(current_path.attribute(_fork_eliminations[key]))
                        current_branch_attrs |= {key}
                current_branch_attrs = branch_attrs | current_branch_attrs
                if _group_by > current_branch_attrs:
                    raise ConfigRequiredAttributesNotFoundError(current_path, _group_by - current_branch_attrs)
                next_branches = []
                for key in config_item:
                    if key in _fork_set:
                        next_branches.append((key, config_item[key]))
                for key, _ in next_branches:
                    config_item.pop(key)
                current_index = c.current
                c.inc()
                yield parent_index, current_index, current_path, current_branch_attrs, not next_branches, config_item
                for key, next_branch in next_branches:
                    next_path = current_path.attribute(key)
                    next_branch_fork_set = branch_fork_set | frozenset([key])
                    yield from _process_config_data(current_index, next_path, next_branch_fork_set,
                                                    current_branch_attrs, next_branch, c)
            else:
                raise ConfigElementTypeError(current_path, dict, type(config_item))
    else:
        raise ConfigElementTypeError(path, list, type(config_data))


def _combine_for_total(branch: FrozenSet[FrozenSet[str]], flat_branch: FrozenSet[str], total: FrozenSet[str],
                       group: FrozenSet[str], remaining_parts: Sequence[FrozenSet[str]]) -> \
        Iterator[FrozenSet[FrozenSet[str]]]:
    if flat_branch == total:
        yield branch
    else:
        feasible_remaining = []
        for re in remaining_parts:
            if not flat_branch.isdisjoint(re) and flat_branch & re > group:
                continue
            feasible_remaining.append(re)
        for n, re in enumerate(feasible_remaining):
            nb = branch | frozenset((re,))
            nfb = flat_branch | re
            yield from _combine_for_total(nb, nfb, total, group, feasible_remaining[n + 1:])


def _enumerate_leaf_index_combinations(combination: MutableSet[FrozenSet[str]],
                                       part_map: Mapping[FrozenSet[str], Set[int]]) -> Iterator[Set[int]]:
    if not combination:
        yield set()
    else:
        parts_key = combination.pop()
        part_ids = part_map[parts_key]
        for part_id in part_ids:
            for next_part_ids in _enumerate_leaf_index_combinations(set(combination), part_map):
                next_part_ids.add(part_id)
                yield next_part_ids


def _walk_branch(index: int, included_indices: Set[int],
                 merged: List[Tuple[Mapping, ConfigPath, Tuple[int, int]]],
                 tree_list: List[Tuple[int, Tuple[Mapping, ConfigPath, Tuple[int, int]]]]) -> None:
    if index in included_indices or index is None:
        return
    else:
        prev_index, data = tree_list[index]
        merged.append(data)
        included_indices.add(index)
        _walk_branch(prev_index, included_indices, merged, tree_list)


def _merge_config_branches(leaf_indices: Set[int],
                           tree_list: List[Tuple[int, Tuple[Mapping, ConfigPath, Tuple[int, int]]]]) -> \
        Tuple[List[Mapping], List[ConfigPath], int]:
    merged = []
    included = set()
    for leaf_index in leaf_indices:
        _walk_branch(leaf_index, included, merged, tree_list)
    merged.sort(key=lambda e: e[2])
    namespaces = []
    paths = []
    for ns, pth, _ in merged:
        namespaces.append(ns)
        paths.append(pth)
    return namespaces, paths, merged[-1][2][1]


def get_app_config_from(config_data) -> AppData:
    config_tree_list = []
    leaf_indices = defaultdict(lambda: defaultdict(set))
    all_branch_attributes = defaultdict(set)
    group_key = None
    for parent_index, current_index, path, branch_attributes, is_leaf, data in \
            _process_config_data(None, ConfigPath('$'), frozenset(), frozenset(), config_data, _Indexer()):
        config_tree_list.append((parent_index, (data, path, (len(path), current_index))))
        if parent_index is None:
            group_key = path
        if is_leaf:
            leaf_indices[group_key][frozenset(branch_attributes)].add(current_index)
            all_branch_attributes[group_key].update(branch_attributes)
    app_entries = []
    app_entry_paths = {}
    for group_key in leaf_indices:
        branch_attributes = frozenset(all_branch_attributes[group_key])
        leaf_indices_by_branch_attribute_sets = leaf_indices[group_key]
        branch_attribute_sets = frozenset(leaf_indices_by_branch_attribute_sets.keys())
        branch_attribute_set_combinations = set(_combine_for_total(frozenset(), frozenset(), branch_attributes,
                                                                   _group_by, list(branch_attribute_sets)))
        not_used_attribute_sets = branch_attribute_sets - {ba_set for ba_set_c in branch_attribute_set_combinations
                                                           for ba_set in ba_set_c}
        if not_used_attribute_sets:
            paths = []
            for attr_set in not_used_attribute_sets:
                for leaf_index in leaf_indices_by_branch_attribute_sets[attr_set]:
                    paths.append(config_tree_list[leaf_index][1][1])
                raise ConfigIncompleteBranchesError(paths, branch_attributes - attr_set)
        for branch_attribute_sets_combination in branch_attribute_set_combinations:
            for leaf_index_combination in \
                    _enumerate_leaf_index_combinations(set(branch_attribute_sets_combination),
                                                       leaf_indices_by_branch_attribute_sets):
                namespaces, paths, sort_key = _merge_config_branches(leaf_index_combination, config_tree_list)
                app_entry = create_app_entry(namespaces, paths)
                entry_key = app_entry.ref
                if entry_key not in app_entry_paths:
                    app_entries.append((app_entry, sort_key))
                    app_entry_paths[entry_key] = paths
                else:
                    raise ConfigDuplicatedEntryIdentityError(app_entry_paths[entry_key], paths, asdict(entry_key))
    app_entries.sort(key=lambda ae: ae[1])
    return AppData([ae[0] for ae in app_entries])


class ConfigProcessor:

    def __init__(self, app_data_reader: AppDataReader):
        self.reader = app_data_reader

    def process_config(self, file_path: str) -> AppData:
        data = self.reader.read(file_path)
        return get_app_config_from(data)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from ctypes import WinDLL, c_wchar_p, WinError
from ctypes.wintypes import HWND, UINT
from winreg import OpenKeyEx, QueryValueEx, CloseKey, ExpandEnvironmentStrings, SetValueEx, HKEY_CURRENT_USER, \
    KEY_SET_VALUE, REG_EXPAND_SZ, HKEY_LOCAL_MACHINE


class PathManager:
    notify_timeout = 1000

    @staticmethod
    def get_system_path() -> str:
        key = OpenKeyEx(HKEY_LOCAL_MACHINE, r'System\CurrentControlSet\Control\Session Manager\Environment')
        try:
            path, _ = QueryValueEx(key, 'Path')
        finally:
            CloseKey(key)
        return path

    @staticmethod
    def get_user_path() -> str:
        key = OpenKeyEx(HKEY_CURRENT_USER, 'Environment')
        try:
            path, _ = QueryValueEx(key, 'Path')
        finally:
            CloseKey(key)
        return path

    @staticmethod
    def extend_path_data(path_entry: str) -> str:
        return ExpandEnvironmentStrings(path_entry)

    @staticmethod
    def set_user_path(path: str) -> None:
        key = OpenKeyEx(HKEY_CURRENT_USER, 'Environment', access=KEY_SET_VALUE)
        try:
            SetValueEx(key, 'Path', 0, REG_EXPAND_SZ, path)
        finally:
            CloseKey(key)

    @staticmethod
    def notify_path_change() -> None:
        send_message_timeout_w = WinDLL('user32', use_last_error=True).SendMessageTimeoutW
        handle = HWND(0xffff)
        message = UINT(0x001A)
        l_param = c_wchar_p('Environment')
        flags = UINT(0x0002 & 0x0008)
        timeout = UINT(__class__.notify_timeout)
        result = send_message_timeout_w(handle, message, None, l_param, flags, timeout, None)
        if result == 0:
            raise WinError()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from dataclasses import dataclass, field
from gettext import translation
from typing import TYPE_CHECKING

from thespiae.conf.exception import AppEntryMixin
from thespiae.exception import ThespiaeError, generate___str__

if TYPE_CHECKING:
    from multidict import CIMultiDictProxy

    from .data import DownloadSpec

t = translation('thespiae', fallback=True)
_ = t.gettext


@generate___str__
@dataclass(frozen=True)
class _DownloadSpecMixin:
    download_spec: DownloadSpec = field(metadata={'format': _('entry for: {0.ref}')})


@generate___str__(_('Invalid or missing configuration data'))
@dataclass(frozen=True)
class InvalidOrMissingAppDataError(AppEntryMixin, ThespiaeError):
    key: str = field(metadata={'format': _('configuration key: {}')})
    value: object = field(metadata={'format': _('current key value: {}')})


@generate___str__(_('Unrecognized application installation type'))
@dataclass(frozen=True)
class UnknownInstallTypeError(AppEntryMixin, ThespiaeError):
    pass


@generate___str__(_('Unrecognized application removal type'))
@dataclass(frozen=True)
class UnknownUninstallTypeError(AppEntryMixin, ThespiaeError):
    pass


@generate___str__(_('Download error'))
@dataclass(frozen=True)
class DownloadError(_DownloadSpecMixin, ThespiaeError):
    pass


@generate___str__(_('File hash does not match the specified one'))
@dataclass(frozen=True)
class HashError(_DownloadSpecMixin, ThespiaeError):
    actual_hash: str = field(metadata={'format': _('actual hash: {}')})


@generate___str__(_('Remote server returned unsupported response'))
@dataclass(frozen=True)
class UnsupportedResponseError(_DownloadSpecMixin, ThespiaeError):
    status: int = field(metadata={'format': _('response status {}')})
    headers: CIMultiDictProxy = field(metadata={'format': _('response headers {}')})


@generate___str__(_('Filesystem operations have been interrupted, please resolve remaining inconsistencies manually \
                     before running the application again'))
@dataclass(frozen=True)
class InterruptedFileOperationsError(AppEntryMixin, ThespiaeError):
    root_directory: str = field(metadata={'format': _('directory: {}')})



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from dataclasses import dataclass, field, fields
from re import finditer
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Tuple

_FORK_ATTR = 'fork_attr'
_GROUP_ATTR = 'group_attr'


class ConfigPath:
    __slots__ = ['_path']

    def __init__(self, path: str = None):
        if path:
            p_end = 0
            parts = []
            for match in finditer(r'(?P<root>\$)|\[(?P<index>\d+)\]|\.(?P<attribute>[A-Za-z]+)', path):
                start, end = match.span()
                if start != p_end:
                    self._path_error(path, p_end)
                root = match.group('root')
                if root:
                    if start == 0:
                        parts.append((None, None))
                    else:
                        self._path_error(path, p_end)
                else:
                    index = match.group('index')
                    if index is not None:
                        parts.append((int(index), None))
                    else:
                        attribute = match.group('attribute')
                        if attribute:
                            parts.append((None, attribute))
                        else:
                            self._path_error(path, p_end)
                p_end = end
            if p_end != len(path):
                self._path_error(path, p_end)
            self._path = tuple(parts)
        else:
            self._path = ()

    @staticmethod
    def _path_error(path: str, pos: int) -> None:
        raise ValueError(f'Invalid path \'{path}\' at position {pos}')

    def index(self, index: int) -> ConfigPath:
        np = __class__()
        np._path = self._path + ((index, None),)
        return np

    def attribute(self, attribute: str) -> ConfigPath:
        np = __class__()
        np._path = self._path + ((None, attribute),)
        return np

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._path == other._path
        return False

    def __hash__(self):
        return hash(self._path)

    def __len__(self):
        return len(self._path)

    def __str__(self):
        items = []
        for index, attribute in self._path:
            if index is not None:
                items.append(f'[{index}]')
            elif attribute is not None:
                items.append(f'.{attribute}')
            else:
                items.append('$')
        return ''.join(items)

    def __repr__(self):
        return f'{__class__.__name__}(\'{self}\')'


@dataclass(frozen=True)
class AppEntryRef:
    name: str = field(metadata={_FORK_ATTR: 'names', _GROUP_ATTR: True})
    version: str = field(default=None, metadata={_FORK_ATTR: 'versions'})
    architecture: str = field(default=None, metadata={_FORK_ATTR: 'architectures'})

    def __str__(self):
        entries = []
        for ref_field in fields(self.__class__):
            if hasattr(self, ref_field.name):
                val = getattr(self, ref_field.name)
                if val is not None:
                    entries.append(str(val))
        return ':'.join(entries)


@dataclass(frozen=True)
class AppEntry(AppEntryRef):
    keep: bool = False
    skip: bool = False
    installer_url: str = None
    package_url: str = None
    uninstaller_path: str = None
    product_code: str = None
    file_hash: str = None
    command: str = None
    install_args: List[str] = None
    uninstall_args: List[str] = None
    list_args: List[str] = None
    installed_list_entry: str = None
    path_entries: List[str] = None
    file_url: str = None
    file_directory: str = None
    file_name: str = None
    archive_url: str = None
    archive_format: str = None
    unpack_directory: str = None

    @property
    def ref(self) -> AppEntryRef:
        return AppEntryRef(self.name, self.version, self.architecture)


class AppData:

    def __init__(self, app_entries: Iterable[AppEntry]):
        self._to_install = tuple((entry for entry in app_entries if not entry.skip and entry.keep))
        self._to_uninstall = tuple((entry for entry in app_entries if not entry.skip and not entry.keep))

    @property
    def to_install(self) -> Tuple[AppEntry, ...]:
        return self._to_install

    @property
    def to_uninstall(self) -> Tuple[AppEntry, ...]:
        return self._to_uninstall
